import shutil
import subprocess as sp
import unittest
from pathlib import Path
from unittest.mock import patch

from click.testing import CliRunner

from ckan_pilot.root import cli


class TestSyncCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        Path("/tmp/ckan-pilot/testprj").mkdir(parents=True)
        Path("/tmp/ckan-pilot/testprj/.ckan-pilot").touch()

    def __init_project(self):
        env_runner = CliRunner()
        setup_test_env_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        env_runner.invoke(cli, setup_test_env_options)

    def test_sync_success(self):
        self.__init_project()
        shutil.rmtree("/tmp/ckan-pilot/testdir/.venv")

        # Sync the project
        test_sync_options = ["-d", "/tmp/ckan-pilot/testdir", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("synchronized", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_sync_on_synced_project(self):
        self.__init_project()

        # Sync the project
        test_sync_options = ["-d", "/tmp/ckan-pilot/testdir", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("synchronized", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_sync_is_not_ckan_pilot_project(self):
        test_sync_options = ["-d", "/tmp", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("not a valid ckan-pilot project", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch.dict("os.environ", {"PATH": "/some/path"}, clear=True)
    def test_sync_tools_missing(self):
        test_sync_options = ["-d", "/tmp/ckan-pilot/testprj", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("not found", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.create_virtual_env",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_sync_fail_create_virtual_env(self, mock_create_virtual_env):
        test_sync_options = ["-d", "/tmp/ckan-pilot/testprj", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("Failed to create a virtual environment", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.setup_ckan_src",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_sync_fail_setup_ckan_src(self, mock_add_ckan_src):
        test_sync_options = ["-d", "/tmp/ckan-pilot/testprj", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("Add CKAN to project failed", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.sync_project",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_sync_fail_sync_project(self, mock_sync_project):
        self.__init_project()
        shutil.rmtree("/tmp/ckan-pilot/testdir/.venv")

        test_sync_options = ["-d", "/tmp/ckan-pilot/testdir", "sync"]
        result = self.runner.invoke(cli, test_sync_options)
        self.assertIn("Failed to sync project", result.output)
        self.assertEqual(result.exit_code, 1)

    def tearDown(self):
        shutil.rmtree("/tmp/ckan-pilot", ignore_errors=True)



import unittest
from unittest.mock import patch

from click.testing import CliRunner

from ckan_pilot.root import cli


class TestVersionCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    @staticmethod
    def _side_effect_except():
        return StopIteration

    @patch("shutil.which", autospec=True)
    def test_tools_list_available(self, mock_which):
        result = self.runner.invoke(cli, ["tools", "list"])
        self.assertIn("available", result.output)
        self.assertEqual(result.exit_code, 0)

    @patch("shutil.which", autospec=True)
    def test_tools_list_not_found(self, mock_which):
        mock_which.return_value = None
        result = self.runner.invoke(cli, ["tools", "list"])
        self.assertIn("not found", result.output)
        self.assertEqual(result.exit_code, 0)

    @patch("shutil.which", autospec=True)
    @patch("platform.system", autospec=True)
    @patch("subprocess.run", autospec=True)
    def test_tools_install_success(self, mock_subprocess_run, mock_system, mock_which):
        mock_which.return_value = None
        mock_system.return_value = "Linux"
        result = self.runner.invoke(cli, ["tools", "install"])
        self.assertIn("Installing", result.output)
        self.assertEqual(result.exit_code, 0)

    @patch("shutil.which", autospec=True)
    @patch("platform.system", autospec=True)
    def test_tools_install_already_installed(self, mock_system, mock_which):
        mock_system.return_value = "Linux"
        result = self.runner.invoke(cli, ["tools", "install"])
        self.assertIn("uv already installed", result.output)
        self.assertEqual(result.exit_code, 0)

    @patch("shutil.which", autospec=True)
    @patch("platform.system", autospec=True)
    def test_tools_install_unsupported_platform(self, mock_which, mock_system):
        mock_which.return_value = None
        mock_system.return_value = "Windows"
        result = self.runner.invoke(cli, ["tools", "install"])
        self.assertIn("Unsupported", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch("urllib.request.urlopen", autospec=True)
    @patch("shutil.which", autospec=True)
    @patch("platform.system", autospec=True)
    @patch("subprocess.run", autospec=True)
    def test_tools_install_exception(self, mock_subprocess_run, mock_system, mock_which, mock_urlopen):
        mock_which.return_value = None
        mock_system.return_value = "Linux"
        mock_subprocess_run.side_effect = self._side_effect_except()
        result = self.runner.invoke(cli, ["tools", "install"])
        self.assertIn("Failed", result.output)
        self.assertEqual(result.exit_code, 1)



import subprocess
import unittest
from unittest.mock import MagicMock, patch

import ckan_pilot.helpers.subproc


class TestHelpers(unittest.TestCase):
    @patch("subprocess.run")
    def test_git_clone_wrapper(self, mock_run):
        """
        Test git_clone_wrapper function.

        Simulates a successful git clone operation and checks that the function
        constructs the correct command and calls subprocess.run with the expected arguments.

        Failure Case:
        - Fails if subprocess.run is not called with the correct arguments.
        """
        # Setup mock
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        # Test parameters
        version = "v1.0"
        git_url = "https://github.com/example/repo"
        extra_flags = ["--depth", "1"]
        clone_target_dir = "/path/to/clone/dir"

        # Call the function
        ckan_pilot.helpers.subproc.git_clone_wrapper(version, git_url, extra_flags, clone_target_dir)

        # Check if subprocess.run was called with the expected command
        mock_run.assert_called_once_with(
            ["git", "clone", "--branch", version, "--depth", "1", git_url, clone_target_dir],
            capture_output=True,
            check=True,
        )

    @patch("subprocess.run")
    def test_git_clone_wrapper_without_extra_flags(self, mock_run):
        """
        Test git_clone_wrapper function when no extra flags are provided.

        Ensures the function correctly constructs the command without additional flags.
        """
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        version = "v1.0"
        git_url = "https://github.com/example/repo"
        extra_flags = None  # No extra flags
        clone_target_dir = "/path/to/clone/dir"

        # Call the function
        ckan_pilot.helpers.subproc.git_clone_wrapper(version, git_url, extra_flags, clone_target_dir)

        # Verify subprocess.run was called with the correct command (no extra flags)
        mock_run.assert_called_once_with(
            ["git", "clone", "--branch", version, git_url, clone_target_dir],  # No extra flags included
            capture_output=True,
            check=True,
        )

    @patch("subprocess.run")
    def test_uv_add_wrapper(self, mock_run):
        """
        Test uv_add_wrapper function.

        Simulates a successful uv add operation and checks that the function
        constructs the correct command and calls subprocess.run with the expected arguments.

        Failure Case:
        - Fails if subprocess.run is not called with the correct arguments.
        """
        # Setup mock
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        # Test parameters
        project_dir = "/path/to/project"
        group = "ckanext"
        target = "extension_name"
        editable = True
        requirements_file = None

        # Call the function
        result = ckan_pilot.helpers.subproc.uv_add_wrapper(project_dir, group, target, editable, requirements_file)

        # Check if subprocess.run was called with the expected command
        mock_run.assert_called_once_with(
            ["uv", "--directory", project_dir, "add", "--group", group, "--editable", target],
            capture_output=True,
            check=True,
        )

        # Ensure the return value is as expected
        self.assertEqual(result.stdout, b"success")
        self.assertEqual(result.stderr, b"")

    @patch("subprocess.run")
    def test_uv_add_wrapper_no_target_or_requirements(self, mock_run):
        """
        Test uv_add_wrapper function with no target or requirements_file, ensuring valid command is run.

        Failure Case:
        - Fails if both target and requirements_file are provided.
        - Fails if no arguments are passed but the function proceeds.
        """
        # Setup mock
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        # Test parameters with no target and no requirements_file
        project_dir = "/path/to/project"
        group = "ckanext"
        target = None
        editable = False
        requirements_file = None

        # Call the function
        result = ckan_pilot.helpers.subproc.uv_add_wrapper(project_dir, group, target, editable, requirements_file)

        # Check if subprocess.run was called with the expected command
        mock_run.assert_called_once_with(
            ["uv", "--directory", project_dir, "add", "--group", group], capture_output=True, check=True
        )

        # Ensure the return value is as expected
        self.assertEqual(result.stdout, b"success")
        self.assertEqual(result.stderr, b"")

    @patch("subprocess.run")
    def test_uv_add_wrapper_with_no_group(self, mock_run):
        """
        Test uv_add_wrapper function when no group is specified.

        Ensures the function correctly includes no --group flag in the command.
        """
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        project_dir = "/path/to/project"
        group = None
        target = "my-package"
        editable = None
        requirements_file = None

        ckan_pilot.helpers.subproc.uv_add_wrapper(project_dir, group, target, editable, requirements_file)

        mock_run.assert_called_once_with(
            ["uv", "--directory", project_dir, "add", target],
            capture_output=True,
            check=True,
        )

    @patch("subprocess.run")
    def test_uv_add_wrapper_with_requirements_file(self, mock_run):
        """
        Test uv_add_wrapper function when a requirements file is provided.

        Ensures the function correctly includes the -r flag with the requirements file.
        """
        mock_run.return_value = MagicMock(stdout=b"success", stderr=b"", returncode=0)

        project_dir = "/path/to/project"
        group = None
        target = None
        editable = None
        requirements_file = "/path/to/requirements.txt"

        ckan_pilot.helpers.subproc.uv_add_wrapper(project_dir, group, target, editable, requirements_file)

        mock_run.assert_called_once_with(
            ["uv", "--directory", project_dir, "add", "-r", requirements_file],
            capture_output=True,
            check=True,
        )

    @patch("subprocess.run")
    def test_uv_remove_wrapper_no_flags(self, mock_run):
        """
        Test case for calling the uv_remove_wrapper function without any group, target, or extra_flags.

        This test simulates a call to the `uv_remove_wrapper` function with no group,
        target, or extra_flags, and verifies that the correct command is constructed and passed to subprocess.run.

        Expected behavior:
        - The function constructs the command with default options and calls `subprocess.run` with the correct arguments
        - The subprocess result is returned correctly.
        """
        # Simulate a subprocess.CompletedProcess return value
        mock_completed_process = MagicMock()
        mock_completed_process.returncode = 0
        mock_completed_process.stderr = ""
        mock_run.return_value = mock_completed_process

        # Call the function with no group, target, or extra_flags
        result = ckan_pilot.helpers.subproc.uv_remove_wrapper("/path/to/project", None, None, None)

        # Check that subprocess.run was called with the correct arguments
        mock_run.assert_called_once_with(
            ["uv", "--directory", "/path/to/project", "remove", "--no-sync"],
            capture_output=True,
            text=True,
            check=False,
        )

        # Assert the return value is the mocked subprocess result
        self.assertEqual(result, mock_completed_process)

    @patch("subprocess.run")
    def test_uv_remove_wrapper_with_group_and_flags(self, mock_run):
        """
        Test case for calling the uv_remove_wrapper function with a group, target, and extra_flags.

        This test simulates a call to the `uv_remove_wrapper` function with a specified group,
        target, and extra_flags, and verifies that the correct command is constructed and passed to subprocess.run.

        Expected behavior:
        - The function constructs the command with the specified group, target, and extra flags.
        - The subprocess result is returned correctly.
        """
        # Simulate a subprocess.CompletedProcess return value
        mock_completed_process = MagicMock()
        mock_completed_process.returncode = 0
        mock_completed_process.stderr = "Error occurred"
        mock_run.return_value = mock_completed_process

        # Call the function with group, target, and extra_flags
        result = ckan_pilot.helpers.subproc.uv_remove_wrapper(
            "/path/to/project", "group-name", "target-name", ["--force", "--quiet"]
        )

        # Check that subprocess.run was called with the correct arguments
        mock_run.assert_called_once_with(
            [
                "uv",
                "--directory",
                "/path/to/project",
                "remove",
                "--no-sync",
                "--group",
                "group-name",
                "--force",
                "--quiet",
                "target-name",
            ],
            capture_output=True,
            text=True,
            check=False,
        )

        # Assert the return value is the mocked subprocess result
        self.assertEqual(result, mock_completed_process)

    @patch("subprocess.run")
    def test_uv_remove_wrapper_with_target_only(self, mock_run):
        """
        Test case for calling the uv_remove_wrapper function with a target only.

        This test simulates a call to the `uv_remove_wrapper` function with only a,
        target specified (and no group or extra_flags), and verifies that the correct,
        command is constructed and passed to subprocess.run.

        Expected behavior:
        - The function constructs the command with the target and default options.
        - The subprocess result is returned correctly.
        """
        # Simulate a subprocess.CompletedProcess return value
        mock_completed_process = MagicMock()
        mock_completed_process.returncode = 0
        mock_completed_process.stderr = ""
        mock_run.return_value = mock_completed_process

        # Call the function with only a target and no group or extra_flags
        result = ckan_pilot.helpers.subproc.uv_remove_wrapper("/path/to/project", None, "target-name", None)

        # Check that subprocess.run was called with the correct arguments
        mock_run.assert_called_once_with(
            ["uv", "--directory", "/path/to/project", "remove", "--no-sync", "target-name"],
            capture_output=True,
            text=True,
            check=False,
        )

        # Assert the return value is the mocked subprocess result
        self.assertEqual(result, mock_completed_process)

    @patch("subprocess.run")
    def test_uv_remove_wrapper_error_handling(self, mock_run):
        """
        Test case for handling errors in the uv_remove_wrapper function.

        This test simulates a scenario where an error occurs during the execution of the command
        (e.g., subprocess fails). It verifies that the function handles errors correctly and,
        that the subprocess result with an error is returned.

        Expected behavior:
        - The function constructs the command with the provided parameters.
        - The subprocess result with an error is returned.
        """
        # Simulate a subprocess.CompletedProcess return value with an error
        mock_completed_process = MagicMock()
        mock_completed_process.returncode = 1
        mock_completed_process.stderr = "Some error occurred"
        mock_run.return_value = mock_completed_process

        # Call the function with parameters that cause an error
        result = ckan_pilot.helpers.subproc.uv_remove_wrapper(
            "/path/to/project", "group-name", "target-name", ["--force"]
        )

        # Check that subprocess.run was called with the correct arguments
        mock_run.assert_called_once_with(
            [
                "uv",
                "--directory",
                "/path/to/project",
                "remove",
                "--no-sync",
                "--group",
                "group-name",
                "--force",
                "target-name",
            ],
            capture_output=True,
            text=True,
            check=False,
        )

        # Assert the return value is the mocked subprocess result
        self.assertEqual(result, mock_completed_process)

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_ckan_cli_base_command(self, mock_run, mock_logger):
        mock_run.return_value = MagicMock(stdout=b"CKAN started", returncode=0)

        ckan_pilot.helpers.subproc.ckan_cli("/fake/project", generate_config=False, dev=False, prod=False)

        mock_run.assert_called_once_with(["uv", "run", "ckan"], capture_output=True, check=True, cwd="/fake/project")
        mock_logger.info.assert_called_with("CKAN started")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_ckan_cli_generate_config(self, mock_run, mock_logger):
        mock_run.return_value = MagicMock(stdout=b"Config generated")

        ckan_pilot.helpers.subproc.ckan_cli("/project", generate_config=True, dev=False, prod=False)

        mock_run.assert_called_once_with(
            ["uv", "run", "ckan", "generate", "config"], capture_output=True, check=True, cwd="/project"
        )
        mock_logger.info.assert_called_with("Config generated")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_ckan_cli_dev_and_prod(self, mock_run, mock_logger):
        mock_run.return_value = MagicMock(stdout=b"Both ini files used")

        ckan_pilot.helpers.subproc.ckan_cli("/project", generate_config=False, dev=True, prod=True)

        mock_run.assert_called_once_with(
            ["uv", "run", "ckan", "development.ini", "production.ini"], capture_output=True, check=True, cwd="/project"
        )
        mock_logger.info.assert_called_with("Both ini files used")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run", side_effect=FileNotFoundError)
    def test_ckan_cli_command_not_found(self, mock_run, mock_logger):
        ckan_pilot.helpers.subproc.ckan_cli("/project", generate_config=False, dev=False, prod=False)

        mock_logger.error.assert_called_with("Command not found: %s", "uv")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run", side_effect=Exception("Boom"))
    def test_ckan_cli_unexpected_error(self, mock_run, mock_logger):
        ckan_pilot.helpers.subproc.ckan_cli("/project", generate_config=False, dev=False, prod=False)

        mock_logger.exception.assert_called_with("An unexpected error occurred: %s", "Boom")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run", side_effect=MagicMock(side_effect=RuntimeError("CKAN failed")))
    def test_ckan_cli_called_process_error(self, mock_run, mock_logger):
        # Simulate a CalledProcessError
        error = subprocess.CalledProcessError(
            returncode=1, cmd=["uv", "run", "ckan"], output=b"Partial output", stderr=b"Something broke"
        )
        mock_run.side_effect = error

        ckan_pilot.helpers.subproc.ckan_cli("/project", generate_config=False, dev=False, prod=False)

        mock_logger.error.assert_any_call("CKAN command failed with exit code %s", 1)
        mock_logger.error.assert_any_call("Stdout:\n%s", "Partial output")
        mock_logger.error.assert_any_call("Stderr:\n%s", "Something broke")

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_start_compose_detached_with_build(self, mock_run, mock_logger):
        mock_run.return_value = MagicMock(returncode=0, stdout="Started in detached mode")
        code = ckan_pilot.helpers.subproc.start_compose("/fake/project", d=True, v=False, b=True, project_name="test")

        mock_run.assert_called_once_with(
            ["docker", "compose", "-p", "test", "up", "--build", "-d"],
            cwd="/fake/project/compose-dev",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            check=False,
        )
        mock_logger.debug.assert_any_call("Started in detached mode")
        self.assertEqual(code, 0)

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_start_compose_detached_without_build(self, mock_run, mock_logger):
        mock_run.return_value = MagicMock(returncode=0, stdout="Detached no build")
        code = ckan_pilot.helpers.subproc.start_compose("/fake/project", d=True, v=False, b=False, project_name="test")

        mock_run.assert_called_once_with(
            ["docker", "compose", "-p", "test", "up", "-d"],
            cwd="/fake/project/compose-dev",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            check=False,
        )
        self.assertEqual(code, 0)

    @patch("ckan_pilot.helpers.subproc._compose_down")
    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.Popen")
    def test_start_compose_foreground_completes(self, mock_popen, mock_logger, mock_down):
        mock_process = MagicMock()
        mock_process.stdout = ["Line 1\n", "Line 2\n"]
        mock_process.wait.return_value = 123
        mock_popen.return_value = mock_process

        code = ckan_pilot.helpers.subproc.start_compose("/fake/project", d=False, v=False, b=False, project_name="test")

        mock_popen.assert_called_once_with(
            ["docker", "compose", "-p", "test", "up", "-w"],
            cwd="/fake/project/compose-dev",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
        )
        mock_logger.debug.assert_any_call("Line 1")
        mock_logger.debug.assert_any_call("Line 2")
        self.assertEqual(code, 123)

    @patch("ckan_pilot.helpers.subproc._compose_down")
    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.Popen")
    def test_start_compose_foreground_keyboard_interrupt(self, mock_popen, mock_logger, mock_down):
        mock_process = MagicMock()

        def side_effect():
            raise KeyboardInterrupt

        mock_process.stdout = MagicMock()
        mock_process.stdout.__iter__.side_effect = side_effect
        mock_process.wait.return_value = 1
        mock_popen.return_value = mock_process

        code = ckan_pilot.helpers.subproc.start_compose("/fake/project", d=False, v=True, b=False, project_name="test")

        mock_down.assert_called_once_with("/fake/project", True, "test")
        self.assertEqual(code, 1)

    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_compose_down_no_volumes(self, mock_run):
        mock_run.return_value = MagicMock(stdout="Stopped containers", stderr="", returncode=0)

        ckan_pilot.helpers.subproc._compose_down("/fake/project", v=False, project_name="testproject")

        mock_run.assert_called_once_with(
            ["docker", "compose", "-p", "testproject", "down"],
            cwd="/fake/project/compose-dev",
            capture_output=True,
            text=True,
            check=False,  # noqa: E501
        )

    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_compose_down_with_volumes(self, mock_run):
        mock_run.return_value = MagicMock(stdout="", stderr="Warning: volume will be removed", returncode=0)

        ckan_pilot.helpers.subproc._compose_down("/fake/project", v=True, project_name="testproject")

        mock_run.assert_called_once_with(
            ["docker", "compose", "-p", "testproject", "down", "-v"],
            cwd="/fake/project/compose-dev",
            capture_output=True,
            text=True,
            check=False,
        )

    @patch("ckan_pilot.helpers.subproc.logger")
    @patch("ckan_pilot.helpers.subproc.sp.run", side_effect=Exception("Docker not found"))
    def test_compose_down_exception(self, mock_run, mock_logger):
        ckan_pilot.helpers.subproc._compose_down("/fake/project", v=True, project_name="testproject")

        mock_logger.error.assert_called_once()
        self.assertIn("Docker not found", mock_logger.error.call_args[0][0])

    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_git_wrapper_success_no_flags(self, mock_run):
        mock_result = MagicMock()
        mock_result.stderr.decode.return_value = ""
        mock_result.stdout = b"Success"
        mock_run.return_value = mock_result

        result = ckan_pilot.helpers.subproc.git_wrapper(["status"], None)
        mock_run.assert_called_once_with(["git", "status"], capture_output=True, check=True)
        self.assertEqual(result, mock_result)

    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_git_wrapper_success_with_flags(self, mock_run):
        mock_result = MagicMock()
        mock_result.stderr.decode.return_value = ""
        mock_result.stdout = b"Success"
        mock_run.return_value = mock_result

        result = ckan_pilot.helpers.subproc.git_wrapper(["log"], "--oneline")
        mock_run.assert_called_once_with(["git", "log", "--oneline"], capture_output=True, check=True)
        self.assertEqual(result, mock_result)

    @patch("ckan_pilot.helpers.subproc.sp.run")
    def test_git_wrapper_failure(self, mock_run):
        error = subprocess.CalledProcessError(returncode=1, cmd=["git", "pull"], stderr=b"fatal: not a git repository")
        mock_run.side_effect = error

        with self.assertRaises(subprocess.CalledProcessError):
            ckan_pilot.helpers.subproc.git_wrapper(["pull"], None)

        mock_run.assert_called_once_with(["git", "pull"], capture_output=True, check=True)



import os
import re

import click
import yaml

from ckan_pilot.root import logger


def extract_repo_name(repo_url):
    # Strip whitespace and trailing slash only
    repo_url = repo_url.strip().rstrip("/")

    # Match the last path component, optionally ending in .git
    match = re.search(r"[:/](?P<name>[^/]+?)(?:\.git)?$", repo_url)
    if match:
        return match.group("name")
    else:
        raise ValueError("Invalid Git URL: {0}".format(repo_url))


def write_local_config_declaration(project_dir, extension_name, config_path):
    """
    Load a local config_declaration.yaml file, parse it, and write to an INI-style config file.

    Args:
        project_dir (str): Root directory of the CKAN project.
        extension_name (str): The name of the extension (e.g. "ckanext-pages").
        config_path (str): Full path to the local config_declaration.yaml file.
    """
    if not os.path.exists(config_path):
        logger.info("Config declaration file not found: {0}".format(config_path))
        logger.info("Please add your own configuration options to .env")
        return

    try:
        with open(config_path, "r") as f:
            declaration_data = yaml.safe_load(f)
    except Exception as err:
        logger.error("Failed to load YAML from {0}".format(config_path))
        raise click.Abort() from err

    try:
        from .catalog import parse_config_declaration  # or import it at the top if same module

        config_output = parse_config_declaration(declaration_data)

        output_file = os.path.join(project_dir, "config", f"{extension_name}.ini")

        os.makedirs(os.path.dirname(output_file), exist_ok=True)

        with open(output_file, "w") as f:
            f.write("### {0} ###\n".format(extension_name))
            f.write(config_output)

        logger.info("Local configuration written to {0}".format(output_file))
    except Exception as err:
        logger.error("Failed to write local config for: {0}".format(extension_name))
        raise click.Abort() from err



import unittest
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from ckan_pilot.commands.catalog import CKAN_EXT_CATALOG_URL, cli


class TestCatalogCommand(unittest.TestCase):
    def setUp(self):
        """Set up test data before each test."""
        # Sample mock YAML data that simulates extensions
        self.mock_yaml_data = {
            "extensions": [
                {"name": "test-extension-1", "description": "A sample extension."},
                {"name": "test-extension-2", "description": "Another extension."},
            ]
        }
        self.runner = CliRunner()  # CLI test runner

        # Define a mock project object that can be used in tests
        self.mock_project = MagicMock()

        self.mock_project.projectdir = "/mock/project"

    @patch("ckan_pilot.helpers.catalog.get")
    @patch("rich.table.Table.add_row")
    def test_list_extensions_with_data(self, mock_add_row, mock_catalog_get):
        """
        Test the 'list' subcommand of the 'extensions' CLI group when called with a valid
        catalog source containing multiple extensions.

        This test mocks the catalog YAML retrieval to provide a controlled set of extensions,
        and patches the Table.add_row method to track how the CLI formats and adds extension
        data to the output table.

        The CLI is invoked with the `--catalog-source` option followed by the list subcommand,
        and a mock project object is passed as the Click context.

        Assertions:
        - The CLI exits successfully (exit code 0).
        - The Table.add_row method is called with correct extension name and description pairs,
        matching the mock YAML data.

        Args:
            mock_add_row (MagicMock): Mock of the rich.table.Table.add_row method.
            mock_catalog_get (MagicMock): Mock of the catalog.get function.

        Returns:
            None
        """
        # Setup mock catalog.get() to return sample data
        mock_catalog_get.return_value = self.mock_yaml_data

        # Prepare CLI context object (project) with projectdir
        ctx_obj = self.mock_project

        # Invoke the CLI as: ckan-pilot extensions list --catalog-source <url>
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "list"],
            obj=ctx_obj,
        )

        # Assert the command succeeded
        self.assertEqual(result.exit_code, 0)

        # Now check the table.add_row calls for extension names and descriptions
        expected_calls = [
            ("test-extension-1", "A sample extension."),
            ("test-extension-2", "Another extension."),
        ]

        actual_calls = [call.args for call in mock_add_row.call_args_list]

        for expected_call in expected_calls:
            self.assertIn(expected_call, actual_calls)

    @patch("ckan_pilot.helpers.catalog.get")  # Mock Catalog YAML loading
    @patch("ckan_pilot.commands.catalog.print")  # Mock print to capture output
    def test_list_extensions_with_empty_data(self, mock_print, mock_catalog_file):
        """
        Test the 'list' command of the extensions CLI when YAML is empty.

        Simulates an empty YAML file. Verifies that no rows are added to the table
        and that an appropriate error message is printed.

        Failure Case:
        - Fails if the table is populated despite no data in the YAML file.
        - Fails if the "No extensions found" message is not printed.
        """
        # Simulate empty YAML data
        mock_catalog_file.return_value = {}

        # Invoke the list_extensions command with the --catalog-source argument
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "list"],
            obj=self.mock_project,
        )

        # Check if the command ran successfully
        self.assertEqual(result.exit_code, 0)

        # Verify that no extensions found was logged
        self.assertIn("No extensions found", result.output)  # Ensure appropriate message is printed

    @patch("ckan_pilot.commands.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.get")  # Mock Catalog YAML loading
    def test_list_extensions_with_invalid_yaml(self, mock_catalog_get, mock_logger):
        """
        Test the 'list' command of the extensions CLI when YAML loading fails.

        Simulates a failure in loading the YAML file (e.g., it returns None).
        Verifies that no rows are added and an error message is logged.
        """
        mock_catalog_get.return_value = None  # Simulate invalid YAML

        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "list"],
            obj=self.mock_project,
        )

        self.assertEqual(result.exit_code, 0)

        mock_logger.error.assert_called_once_with("No extensions found in catalog: {0}".format(CKAN_EXT_CATALOG_URL))

    @patch("ckan_pilot.commands.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.helpers.catalog.install_extension")
    @patch("ckan_pilot.helpers.catalog.install_requirements_file")
    @patch("ckan_pilot.helpers.catalog.write_configuration_file")
    def test_extension_add(
        self,
        mock_write_configuration_file,
        mock_install_requirements_file,
        mock_install_extension,
        mock_fetch_extension_metadata,
        mock_logger,
    ):
        """
        Test the 'add' command of the extensions CLI for adding an extension from the catalog.

        Verifies that the extension is added successfully by checking if:
        - The extension metadata is fetched.
        - The extension is installed using the correct URL and version.
        - The required dependencies are installed.

        Failure Case:
        - Fails if the extension installation or requirement installations fail.
        - Fails if the configuration file is not written correctly.
        """
        # Define mock return values
        mock_extension_data = {
            "url": "https://github.com/ckan/ckanext-xloader",
            "version": "2.1.0",
            "setup": {
                "required_system_packages": [],
                "has_requirements": True,
                "has_dev_requirements": False,
            },
        }
        mock_fetch_extension_metadata.return_value = mock_extension_data

        # Run the CLI command with the runner
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "add", "ckanext-xloader"],
            obj=self.mock_project,
        )

        # Ensure the exit status is correct
        self.assertEqual(result.exit_code, 0)

        # Assert that the correct functions were called
        mock_fetch_extension_metadata.assert_called_once_with(
            "https://extensions.ckan.app/catalog.yaml", "ckanext-xloader"
        )
        mock_install_extension.assert_called_once_with(
            self.mock_project.projectdir, "https://github.com/ckan/ckanext-xloader", "2.1.0", False, "ckanext-xloader"
        )
        mock_install_requirements_file.assert_called_once_with(
            self.mock_project.projectdir,
            "https://github.com/ckan/ckanext-xloader",
            "2.1.0",
            "ckanext-xloader",
            "requirements.txt",
            editable=False,
        )
        mock_write_configuration_file.assert_called_once_with(
            self.mock_project.projectdir, mock_extension_data, "ckanext-xloader"
        )
        mock_logger.info.assert_any_call("Installing requirements for: ckanext-xloader")

    @patch("ckan_pilot.commands.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.helpers.catalog.install_extension")
    @patch("ckan_pilot.helpers.catalog.install_requirements_file")
    @patch("ckan_pilot.helpers.catalog.write_configuration_file")
    def test_extension_add_dev_mode(
        self,
        mock_write_configuration_file,
        mock_install_requirements_file,
        mock_install_extension,
        mock_fetch_extension_metadata,
        mock_logger,
    ):
        # Mock extension data to simulate a proper extension with requirements
        mock_extension_data = {
            "url": "https://github.com/ckan/ckanext-xloader",
            "version": "2.1.0",
            "setup": {
                "required_system_packages": [],
                "has_requirements": True,
                "has_dev_requirements": True,
            },
        }
        mock_fetch_extension_metadata.return_value = mock_extension_data

        # Simulate running the CLI command with --dev flag
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "add", "--dev", "ckanext-xloader"],
            obj=self.mock_project,
        )

        self.assertEqual(result.exit_code, 0)

        # Assert the correct fetch and install functions were called
        mock_fetch_extension_metadata.assert_called_once_with(
            "https://extensions.ckan.app/catalog.yaml", "ckanext-xloader"
        )
        mock_install_extension.assert_called_once_with(
            self.mock_project.projectdir, "https://github.com/ckan/ckanext-xloader", "2.1.0", True, "ckanext-xloader"
        )

        # Assert the install_requirements_file was called for both sets of requirements
        mock_install_requirements_file.assert_any_call(
            self.mock_project.projectdir,
            "https://github.com/ckan/ckanext-xloader",
            "2.1.0",
            "ckanext-xloader",
            "requirements.txt",
            editable=False,
        )
        mock_install_requirements_file.assert_any_call(
            self.mock_project.projectdir,
            "https://github.com/ckan/ckanext-xloader",
            "2.1.0",
            "ckanext-xloader-dev",
            "dev-requirements.txt",
            editable=False,
        )

        # Check that logger.info was called for both requirement installations
        mock_logger.info.assert_any_call("Installing requirements for: ckanext-xloader")
        mock_logger.info.assert_any_call("Installing dev requirements for: ckanext-xloader")

        # Check that write_configuration_file was called to save the extension's configuration
        mock_write_configuration_file.assert_called_once_with(
            self.mock_project.projectdir, mock_extension_data, "ckanext-xloader-dev"
        )

    @patch("ckan_pilot.commands.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.helpers.catalog.install_extension")
    @patch("ckan_pilot.helpers.catalog.install_requirements_file")
    @patch("ckan_pilot.helpers.catalog.write_configuration_file")
    def test_extension_add_requirements_only(
        self,
        mock_write_configuration_file,
        mock_install_requirements_file,
        mock_install_extension,
        mock_fetch_extension_metadata,
        mock_logger,
    ):
        # Define mock return values
        mock_extension_data = {
            "url": "https://github.com/ckan/ckanext-someext",
            "version": "1.0.0",
            "setup": {
                "required_system_packages": [],
                "has_requirements": True,
                "has_dev_requirements": False,
            },
        }
        mock_fetch_extension_metadata.return_value = mock_extension_data

        # Mock install_extension to simulate success
        mock_install_extension.return_value = None  # Simulate successful extension installation

        # Run the CLI command
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "add", "ckanext-someext"],
            obj=self.mock_project,
        )

        # Assert the CLI command succeeded
        assert result.exit_code == 0

        # Assert the correct functions were called
        mock_fetch_extension_metadata.assert_called_once_with(
            "https://extensions.ckan.app/catalog.yaml", "ckanext-someext"
        )
        mock_install_extension.assert_called_once_with(
            self.mock_project.projectdir, "https://github.com/ckan/ckanext-someext", "1.0.0", False, "ckanext-someext"
        )
        mock_install_requirements_file.assert_called_once_with(
            self.mock_project.projectdir,
            "https://github.com/ckan/ckanext-someext",
            "1.0.0",
            "ckanext-someext",
            "requirements.txt",
            editable=False,
        )
        mock_logger.info.assert_any_call("Installing requirements for: ckanext-someext")

        mock_write_configuration_file.assert_called_once_with(
            self.mock_project.projectdir, mock_extension_data, "ckanext-someext"
        )

    @patch("ckan_pilot.commands.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.helpers.catalog.install_extension")
    @patch("ckan_pilot.helpers.catalog.install_requirements_file")
    @patch("ckan_pilot.helpers.catalog.write_configuration_file")
    def test_extension_add_no_requirements(
        self,
        mock_write_configuration_file,
        mock_install_requirements_file,
        mock_install_extension,
        mock_fetch_extension_metadata,
        mock_logger,
    ):
        """
        Test the 'add' command of the extensions CLI for an extension that has no requirements.
        This verifies that the function correctly skips the requirement installation.
        """
        # Define mock return values for an extension with no requirements
        mock_extension_data = {
            "url": "https://github.com/ckan/ckanext-xloader",
            "version": "2.1.0",
            "setup": {
                "required_system_packages": [],
                "has_requirements": False,
                "has_dev_requirements": False,
            },
        }
        mock_fetch_extension_metadata.return_value = mock_extension_data

        # Run the CLI command
        result = self.runner.invoke(
            cli,
            ["--catalog-source", CKAN_EXT_CATALOG_URL, "add", "ckanext-xloader"],
            obj=self.mock_project,
        )

        # Ensure the exit status is correct
        self.assertEqual(result.exit_code, 0)

        # Assert the correct functions were called
        mock_fetch_extension_metadata.assert_called_once_with(
            "https://extensions.ckan.app/catalog.yaml", "ckanext-xloader"
        )
        mock_install_extension.assert_called_once_with(
            self.mock_project.projectdir, "https://github.com/ckan/ckanext-xloader", "2.1.0", False, "ckanext-xloader"
        )

        # Ensure that no requirement file installation occurs
        mock_install_requirements_file.assert_not_called()

        # Check that write_configuration_file was called to save the extension's configuration
        mock_write_configuration_file.assert_called_once_with(
            self.mock_project.projectdir, mock_extension_data, "ckanext-xloader"
        )

    @patch("ckan_pilot.helpers.catalog.fetch_extension_metadata")
    def test_invalid_catalog_source(self, mock_fetch_extension_metadata):
        """
        Test what happens when the catalog source URL is invalid.

        This test simulates an invalid catalog source URL and verifies that:
        - The correct exception is raised.
        - The exit code is non-zero to indicate failure.
        - The exception message is correctly included in the result output.

        In this case, we simulate an invalid catalog source by making the
        `fetch_extension_metadata` function raise an exception.
        """
        # Simulate the exception when the fetch_extension_metadata is called
        mock_fetch_extension_metadata.side_effect = Exception("Invalid catalog source")

        # Run the CLI command
        result = self.runner.invoke(
            cli,
            ["--catalog-source", "invalid_url", "add", "ckanext-xloader"],
            obj=self.mock_project,
        )

        # Check that the result exit code is not 0 (indicating failure)
        self.assertNotEqual(result.exit_code, 0)

        # Check if the exception is in the output or captured via exception info
        if result.exc_info:
            exception_type, exception_value, traceback = result.exc_info
            self.assertIn("Invalid catalog source", str(exception_value))



DOCKERFILE_CONTENT = """# ---------- Build Stage ----------
FROM alpine:3.22 AS build

# Set uv config
ENV XDG_BIN_HOME="/usr/local/bin"
ENV UV_COMPILE_BYTECODE="1"
ENV UV_LINK_MODE="copy"
ENV UV_NO_CACHE="1"
ENV UV_PROJECT_ENVIRONMENT="/opt/uv/venv"

# Set the default app dir
ENV APP_DIR="/app"

# Set the location of the virtual environment
ENV PATH="${UV_PROJECT_ENVIRONMENT}/bin:$PATH"

# Set default shell to ash and enable pipefail
SHELL ["/bin/ash", "-o", "pipefail", "-c"]

# System packages to build CKAN requirements and plugins
RUN apk add --no-cache \\
    git>=2.49 \\
    curl>=8.14 \\
    ca-certificates \\
    clang>=20.1 \\
    gcc>=14.2 \\
    linux-headers>=6.14 \\
    musl-dev>=1.2 \\
    libmagic>=5.46 \\
    postgresql17-client>=17.6 \\
    postgresql17-dev>=17.6 && \\
    ln -s /usr/bin/clang /usr/bin/musl-clang

# Install uv
RUN curl --retry 5 --retry-delay 3 --proto '=https' --tlsv1.2 -Ls https://github.com/astral-sh/uv/releases/download/0.7.22/uv-installer.sh | sh

WORKDIR ${APP_DIR}

# Copy the pyproject.toml
COPY pyproject.toml ${APP_DIR}/pyproject.toml

# Initalize venv
RUN uv venv

# Copy CKAN required files and directories
COPY src/ckan /app/src/ckan
COPY extensions /app/extensions

# Install uwsgi
RUN uv add uwsgi

# Sync CKAN pilot project
RUN uv sync --all-groups

# ---------- Final Stage ----------
FROM alpine:3.22 AS final

# Set uv config
ENV XDG_BIN_HOME="/usr/local/bin"
ENV UV_COMPILE_BYTECODE="1"
ENV UV_LINK_MODE="copy"
ENV UV_NO_CACHE="1"
ENV UV_PROJECT_ENVIRONMENT="/opt/uv/venv"

# Set the default app dir
ENV APP_DIR="/app"
WORKDIR ${APP_DIR}

# Set the location of the virtual environment
ENV PATH="${UV_PROJECT_ENVIRONMENT}/bin:$PATH"

# Set default shell to ash and enable pipefail
SHELL ["/bin/ash", "-o", "pipefail", "-c"]

# Copy venv from build stage and other requirements
COPY --from=build /opt/uv/venv /opt/uv/venv
COPY --from=build ${APP_DIR}/uv.lock ${APP_DIR}/uv.lock
COPY --from=build ${APP_DIR}/pyproject.toml ${APP_DIR}/pyproject.toml
COPY --from=build ${APP_DIR}/extensions ${APP_DIR}/extensions
COPY --from=build ${APP_DIR}/src/ckan ${APP_DIR}/src/ckan
# Copy .python-version so we can install correct version in new stage
COPY .python-version ${APP_DIR}/.python-version
# Copy uv binaries
COPY --from=build /usr/local/bin/uv /usr/local/bin/uv

# Install necessary packages to run CKAN
RUN apk add --no-cache \\
        git>=2.49 \\
        curl>=8.14 \\
        ca-certificates \\
        # # Required for CKAN
        libmagic>=5.46 \\
        postgresql17-client>=17.6

# Set python version var and install python
RUN uv python install $(cat /app/.python-version)

# Sync CKAN pilot project
RUN uv sync --all-groups

# Copy CKAN generated ini
{ckan_ini}

# Install additional system packages
{additional_system_packages}

# Copy prerun script as entrypoint and set entrypoint
COPY compose-dev/services/ckan/image/ /app/src/ckan
ENTRYPOINT [ "sh", "/app/src/ckan/docker-entrypoint.sh" ]

CMD {cmd}"""



import os

import click
import docker

from ckan_pilot.helpers import catalog, subproc
from ckan_pilot.root import CKAN_EXT_CATALOG_URL, logger
from ckan_pilot.templates import docker as docker_template


@click.group("dist", short_help="Manages distributions")
@click.option(
    "--catalog-source",
    default=CKAN_EXT_CATALOG_URL,
    help="Path to CKAN extension catalog file or URL, default: https://extensions.ckan.app/catalog.yaml",
)
def cli(catalog_source):
    """
    Group of commands related to managing extensions.
    """
    # Store the global catalog in a context variable
    logger.debug("Using Catalog source: {0}".format(catalog))
    cli.catalog_source = catalog_source


@cli.command("build", short_help="Build CKAN Docker image")
@click.option(
    "--tag",
    help="Tag name for the Docker image",
)
@click.option(
    "--dev",
    help="If specified starts CKAN for local-dev",
    is_flag=True,
)
@click.pass_obj
def build(project, tag, dev):  # noqa: PLR0912 PLR0915
    """
    Build a Docker image for the CKAN project using a dynamically generated Dockerfile.

    The image is built using the local CKAN project directory and the Dockerfile generated
    based on whether the development mode is enabled.

    Args:
        project (object): The CKAN project object passed by Click. It should contain `.projectdir`
                          and `.get_project_metadata()` methods.
        tag (str): Optional. The Docker image tag to use (e.g., 'my-ckan:2.10'). If not provided,
                   it is generated from the project's metadata.
        dev (str): Optional. If specified, builds the image for local development mode.

    Raises:
        click.ClickException: If Docker build or API operations fail.
    """

    if not tag:
        tag = "{0}".format(project.get_project_metadata()[0])

    logger.debug("Setting docker image tag to {0}".format(tag))

    # Get projectdir
    projectdir = project.projectdir

    # Use the global catalog from the context
    catalog_source = cli.catalog_source

    # Dockerfile that is generated and values that are replaced
    dockerfile_content = docker_template.DOCKERFILE_CONTENT

    excluded_extensions = ["ckanext-envvars"]
    installed_extensions_in_toml = catalog.get_installed_extensions(projectdir + "/pyproject.toml", excluded_extensions)

    system_requirements = []
    before_init = []

    for ext in installed_extensions_in_toml:
        data = catalog.fetch_extension_metadata(catalog_source, ext)

        # Skip if no data is returned
        if not data:
            continue

        # --- Collect system requirements ---
        dist_data = data.get("setup", {}).get("distributions", {})
        if isinstance(dist_data, dict):
            alpine_data = dist_data.get("alpine", {})
            if isinstance(alpine_data, dict):
                system_requirements += alpine_data.get("required_system_packages", [])

        # --- Collect init-config commands ---
        init_config = data.get("setup", {}).get("init-config", [])
        if isinstance(init_config, list):
            before_init.extend(init_config)

    # Filter out duplicates
    system_requirements = list(set(system_requirements))

    # Add system requriements to dockerfile
    dockerfile_content = add_system_requirements(system_requirements, dockerfile_content)

    # Generate Dockerfile content based on mode
    if dev:
        # Generate dev ini
        logger.info("Generating dev ini...")
        subproc.ckan_cli(project.projectdir, generate_config=True, dev=True, prod=False)
        # Add envvars to ckan.plugins in ini
        catalog.set_ini_key_value_preserve_comments(
            project.projectdir + "/development.ini", "app:main", "ckan.plugins", "envvars"
        )
        # Set dev ckan.ini
        ckan_ini = "COPY development.ini /app/src/ckan/ckan.ini"
        # Set docker CMD
        cmd = '["ckan", "-c", "/app/src/ckan/ckan.ini", "run", "-p", "5000", "-H", "0.0.0.0"]'
        dockerfile_content = dockerfile_content.replace("{ckan_ini}", ckan_ini).replace("{cmd}", cmd)
    else:
        # Generate prod ini
        logger.info("Generating prod ini...")
        subproc.ckan_cli(project.projectdir, generate_config=True, dev=False, prod=True)
        # Add envvars to ckan.plugins in ini
        catalog.set_ini_key_value_preserve_comments(
            project.projectdir + "/production.ini", "app:main", "ckan.plugins", "envvars"
        )
        # Set prod ckan.ini
        ckan_ini = "COPY production.ini /app/src/ckan/ckan.ini"
        # Set docker CMD
        cmd = '["uwsgi", "--socket", "/tmp/uwsgi.sock", "--http", ":5000", "--master", "--enable-threads", "--wsgi-file", "/app/src/ckan/wsgi.py", "--lazy-apps", "-p", "2", "-L", "--vacuum", "--harakiri", "50", "--callable", "application"]'  # noqa: E501
        dockerfile_content = dockerfile_content.replace("{ckan_ini}", ckan_ini).replace("{cmd}", cmd)

    dockerfile_path = os.path.join(projectdir, "Dockerfile")

    logger.debug("Generating Dockerfile in {0}".format(projectdir))

    # Write the Dockerfile
    with open(dockerfile_path, "w") as dockerfile:
        dockerfile.write(dockerfile_content)

    # Add before-init command to the entrypoint script
    entrypoint_script_path = os.path.join(projectdir, "compose-dev", "services", "ckan", "image", "before-init.sh")
    create_entrypoint_script(entrypoint_script_path, before_init)

    client = docker.APIClient(base_url="unix://var/run/docker.sock")

    logger.debug("Building Docker image '{0}' from {1}".format(tag, projectdir))
    logger.info("Please wait while logs are being generated...")
    try:
        logs = client.build(path=projectdir, tag=tag, rm=True, decode=True, nocache=True)

        for chunk in logs:
            if "stream" in chunk:
                logger.info(chunk["stream"].strip())
            elif "error" in chunk:
                logger.error("Build error: " + chunk["error"].strip())
                raise click.ClickException("Docker build failed: " + chunk["error"].strip())

    except docker.errors.BuildError as e:
        logger.error("Build failed.")
        for err in e.build_log:
            if "stream" in err:
                logger.error(err["stream"].strip())
        raise click.ClickException("Docker build failed.") from e

    except docker.errors.APIError as e:
        logger.error("Docker API error: {0}".format(str(e)))
        logger.error("Docker build failed.")
        raise click.ClickException("Docker build failed.") from e

    logger.info("Image '{0}' built successfully.".format(tag))


def add_system_requirements(system_requirements, dockerfile_content):
    """
    Inserts system package installation commands into a Dockerfile template.

    This function takes a list of system package requirements and appends them
    to a Dockerfile's content by replacing the `{additional_system_packages}`
    placeholder with a properly formatted `RUN apk add --no-cache` command.
    If no requirements are provided, the placeholder is replaced with an empty string.

    Parameters:
        system_requirements (list of str): A list of system packages to install.
        dockerfile_content (str): The content of the Dockerfile containing a
        `{additional_system_packages}` placeholder.

    Returns:
        str: The updated Dockerfile content with system package installation
             commands inserted, or with the placeholder removed if no packages
             are specified.
    """
    if system_requirements:
        logger.info("Writing system requirements in Dockerfile...")

        additional_system_packages = "RUN apk add --no-cache"
        for i, requirement in enumerate(system_requirements):
            if len(system_requirements) == 1:
                additional_system_packages += " {}".format(requirement)
                break

            if i == 0:
                additional_system_packages += " {} \\\n".format(requirement)
            elif i < len(system_requirements) - 1:
                additional_system_packages += "\t" * 2 + "{} \\\n".format(requirement)
            else:
                additional_system_packages += "\t" * 2 + requirement

        return dockerfile_content.replace("{additional_system_packages}", additional_system_packages)
    else:
        return dockerfile_content.replace("{additional_system_packages}", "")


def create_entrypoint_script(output_path, before_init_cmds=None):
    if not before_init_cmds:
        logger.debug("No before-init commands provided, skipping entrypoint script creation.")
        script_lines = [
            "#!/bin/sh",
            "set -e",
            "",
            'echo "Running before-init commands..."',
            "# Example prerun command",
            "# ckan -c /app/src/ckan/ckan.ini db init",
        ]

    # Prepend CKAN command prefix to each command
    before_init_cmds = ["ckan -c /app/src/ckan/ckan.ini " + cmd for cmd in before_init_cmds]

    script_lines = [
        "#!/bin/sh",
        "set -e",
        "",
        'echo "Running before-init commands..."',
        "# Example prerun command",
        "# ckan -c /app/src/ckan/ckan.ini db init",
        "# Run any before-init commands",
    ]

    for cmd in before_init_cmds:
        script_lines.append(cmd)

    with open(output_path, "w") as f:
        f.write("\n".join(script_lines) + "\n")

    # Make the script executable
    os.chmod(output_path, 0o755)

    logger.debug("Entrypoint script created at: {}".format(output_path))



import click

from ckan_pilot.helpers import subproc
from ckan_pilot.root import logger


@click.command()
@click.option("-v", is_flag=True, help="Delete disks on cleanup")
@click.pass_obj
def cli(project, v):
    """Stop the CKAN pilot project."""
    logger.info("Stopping CKAN Pilot...")

    project_name = "{0}".format(project.get_project_metadata()[0])

    subproc._compose_down(project.projectdir, v=v, project_name=project_name)



import subprocess as sp
import traceback
from platform import system
from shutil import which
from urllib.request import urlopen

import click

from ckan_pilot.root import logger


# Command to get the current version
@click.group("tools", short_help="Ensure required tools")
def cli():
    """Ensure required tools"""
    pass


@cli.command("list", short_help="List installed tools")
def list():
    """List installed tools and their versions"""
    uv_path = which("uv")
    if uv_path is None:
        logger.info("uv executable not found on $PATH")
    else:
        get_uv_version = sp.run(["uv", "version"], capture_output=True, check=False)
        uv_version = get_uv_version.stdout.strip().decode("utf-8")
        logger.info("{0} available on PATH at: {1}".format(uv_version, uv_path))


@cli.command("install", short_help="Install tools")
def install():
    """Install tools"""
    uv_path = which("uv")
    os_platform = system()

    if uv_path is None and os_platform in ["Linux", "Darwin"]:
        try:
            logger.info("Installing uv, please follow any additional steps from installation")
            uv_install = urlopen("https://astral.sh/uv/install.sh")
            uv_install_script = uv_install.read().decode("utf-8")
            uv_installation = sp.run([uv_install_script + " | sh"], capture_output=True, shell=True, check=True)
            logger.info(uv_installation.stdout.decode("utf-8"))
        except Exception:
            logger.error("Failed to install uv, traceback below: ")
            logger.error(traceback.format_exc())
            raise click.Abort() from None
    elif uv_path is not None and os_platform in ["Linux", "Darwin"]:
        logger.info("uv already installed: {0}".format(uv_path))
    else:
        logger.error("Unsupported platform: {0}".format(os_platform))
        raise click.Abort()



import os
import subprocess as sp
import unittest
from tempfile import NamedTemporaryFile
from unittest.mock import MagicMock, mock_open, patch

import click
import requests
import tomllib
import yaml

import ckan_pilot.helpers.catalog

# Sample YAML as a Python dictionary
mock_catalog_data = yaml.safe_load("""
extensions:
  - name: "ckanext-xloader"
    url: "https://github.com/ckan/ckanext-xloader"
    version: "2.1.0"
    description: >-
      Loads CSV (and similar) data into CKAN's DataStore.
      Designed as a replacement for DataPusher
      because it offers ten times the speed and more robustness
      (hence the name, derived from "Express Loader").
    supported_ckan_versions:
      - "2.11"
      - "2.10"
    plugins:
      - "xloader"
    configuration:
      has_config_declaration: true
      options:
    setup:
      required_system_packages: [sqlalchemy]
      has_pyproject_toml: true
      has_requirements: true
      has_dev_requirements: true
      init-config:
      afterinit-config:
    metadata:
      keywords:
        - "xloader"
        - "datastore"
        - "loader"
        - "express"

  - name: "ckanext-pages"
    url: "https://github.com/ckan/ckanext-pages"
    version: "v0.5.2"
    description: >-
      This extension gives you an easy way to add simple pages to CKAN.
    supported_ckan_versions:
      - "2.10"
    plugins:
      - "pages"
    configuration:
      has_config_declaration: false
      options:
        - key: "ckanext.pages.organization"
          example: "False"
          default: "False"
          required: false
    setup:
      required_system_packages: []
      has_requirements: true
      has_dev_requirements: true
    metadata:
      keywords:
        - "pages"
        - "ckanext-pages"
""")


# Normalize indentation by removing the leading spaces
def normalize_indentation(text):
    """
    Normalizes the indentation of a given text by removing leading spaces.

    Args:
        text (str): The input text to normalize.

    Returns:
        str: The input text with normalized indentation.
    """
    return "\n".join([line.lstrip() for line in text.strip().split("\n")])


class TestHelpers(unittest.TestCase):
    @patch("requests.get")  # Mocking requests.get for URL retrieval
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_from_url(self, mock_logger, mock_get):
        """
        Test the retrieval of YAML data from a URL.
        """
        # Setup the mock response for requests.get
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.text = yaml.dump({"key": "value"})
        mock_get.return_value = mock_response

        catalog_url = "http://example.com/catalog.yaml"

        # Call the method
        result = ckan_pilot.helpers.catalog.get(catalog_url)

        # Assert the result matches the expected parsed YAML
        self.assertEqual(result, {"key": "value"})

        # Check if requests.get was called correctly
        mock_get.assert_called_once_with(catalog_url, timeout=10)

        # Check if the correct log messages were recorded
        mock_logger.debug.assert_any_call("Fetching YAML from URL: {0}".format(catalog_url))

    @patch("builtins.open", new_callable=mock_open, read_data=yaml.dump({"key": "value"}))
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_from_file(self, mock_logger, mock_file):
        """
        Test the retrieval of YAML data from a local file.
        """
        catalog_file = "path/to/catalog.yaml"

        # Call the method
        result = ckan_pilot.helpers.catalog.get(catalog_file)

        # Assert the result matches the expected parsed YAML
        self.assertEqual(result, {"key": "value"})

        # Check if the open function was called correctly
        mock_file.assert_called_once_with(catalog_file, "r")

        # Check if the correct log messages were recorded
        mock_logger.info.assert_any_call("Loading YAML from file: {0}".format(catalog_file))

    @patch("requests.get")  # Mocking requests.get for failure scenario
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_from_url_failure(self, mock_logger, mock_get):
        """
        Test the failure scenario when retrieving YAML data from a URL.
        """
        # Mock the response to raise an exception
        mock_get.side_effect = requests.exceptions.RequestException("Network error")

        catalog_url = "http://example.com/catalog.yaml"
        with self.assertRaises(click.Abort):  # Expecting click.Abort to be raised, not just Exception
            ckan_pilot.helpers.catalog.get(catalog_url)

        # Check if the correct error log was recorded
        mock_logger.error.assert_any_call("Error fetching YAML from URL: {0}".format(catalog_url))

    @patch("builtins.open")
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_from_file_not_found(self, mock_logger, mock_file):
        """
        Test the failure scenario when the file is not found.
        """
        # Simulate a FileNotFoundError when trying to open the file
        mock_file.side_effect = FileNotFoundError

        catalog_file = "path/to/nonexistent/catalog.yaml"
        with self.assertRaises(click.Abort):  # Expecting click.Abort to be raised, not just Exception
            ckan_pilot.helpers.catalog.get(catalog_file)

        # Check if the correct error log was recorded
        mock_logger.error.assert_any_call("Error: File not found - {0}".format(catalog_file))

    @patch("builtins.open", new_callable=mock_open, read_data="invalid: yaml: data")
    @patch("yaml.safe_load", side_effect=yaml.YAMLError("YAML parsing error"))  # Mock safe_load instead of read
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_from_yaml_error(self, mock_logger, mock_yaml, mock_file):
        """
        Test the failure scenario when the YAML parsing fails.
        """
        catalog_file = "path/to/catalog.yaml"

        with self.assertRaises(click.Abort):  # Expecting click.Abort to be raised
            ckan_pilot.helpers.catalog.get(catalog_file)

        # Check if the correct error log was recorded
        mock_logger.error.assert_any_call(f"Error parsing YAML: {catalog_file}")

    @patch("builtins.open")
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_get_unexpected_error(self, mock_logger, mock_file):
        """
        Test the failure scenario for an unexpected error.
        """
        # Simulate a generic exception
        mock_file.side_effect = Exception("Unexpected error")

        catalog_file = "path/to/catalog.yaml"
        with self.assertRaises(click.Abort):  # Expecting click.Abort to be raised
            ckan_pilot.helpers.catalog.get(catalog_file)

        # Check if the correct error log was recorded
        mock_logger.error.assert_any_call("Unexpected error: {0}".format("Unexpected error"))

    @patch("builtins.open", new_callable=mock_open, read_data="key = 'value'")
    @patch("tomllib.load")
    def test_get_from_toml_success(self, mock_load, mock_open):
        """
        Test case for successful TOML file parsing.

        This test simulates the scenario where a TOML file is successfully loaded and parsed.
        It mocks the `open` and `tomllib.load` functions to verify that the file is opened correctly,
        the `tomllib.load` function is called to parse the file, and the correct result is returned.

        Expected behavior:
        - The file is opened with the correct path.
        - The `tomllib.load` function is called to parse the file.
        - The parsed content is returned correctly.
        """
        # Simulate TOML parsing success
        mock_load.return_value = {"key": "value"}

        # Test the function
        result = ckan_pilot.helpers.catalog.get_from_toml("/path/to/toml/file")

        # Check that the file was opened with the correct path
        mock_open.assert_called_once_with("/path/to/toml/file", "rb")

        # Check that tomllib.load was called to parse the file
        mock_load.assert_called_once_with(mock_open())

        # Assert the return value is correct
        self.assertEqual(result, {"key": "value"})

    @patch("builtins.open", new_callable=mock_open)
    @patch("tomllib.load")
    def test_get_from_toml_file_not_found(self, mock_load, mock_open):
        """
        Test case for handling a FileNotFoundError when loading a TOML file.

        This test simulates the scenario where the TOML file does not exist and a `FileNotFoundError` is raised.
        It verifies that the function raises a `click.Abort` exception as expected when the file is not found.

        Expected behavior:
        - A `click.Abort` exception is raised when the file is not found.
        - The file is opened with the correct path.
        - `tomllib.load` is not called due to the missing file.
        """
        # Simulate a FileNotFoundError
        mock_open.side_effect = FileNotFoundError

        # Test the function and assert that it raises click.Abort()
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.get_from_toml("/path/to/nonexistent/file")

        # Check that the file was opened with the correct path
        mock_open.assert_called_once_with("/path/to/nonexistent/file", "rb")

        # Ensure tomllib.load was not called because the file wasn't found
        mock_load.assert_not_called()

    @patch("builtins.open", new_callable=mock_open, read_data="key = 'value'")
    @patch("tomllib.load")
    def test_get_from_toml_toml_decode_error(self, mock_load, mock_open):
        """
        Test case for handling a TOML decoding error.

        This test simulates the scenario where there is an error while decoding the TOML file.
        It verifies that the function raises a `click.Abort` exception when a `tomllib.TOMLDecodeError` occurs.

        Expected behavior:
        - A `click.Abort` exception is raised due to the TOML decode error.
        - The file is opened with the correct path.
        - The `tomllib.load` function is called and raises the decode error.
        """
        # Simulate a TOML decode error
        mock_load.side_effect = tomllib.TOMLDecodeError("Error in TOML file")

        # Test the function and assert that it raises click.Abort()
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.get_from_toml("/path/to/toml/file")

        # Check that the file was opened with the correct path
        mock_open.assert_called_once_with("/path/to/toml/file", "rb")

        # Ensure tomllib.load was called and it raised the decode error
        mock_load.assert_called_once_with(mock_open())

    @patch("builtins.open", new_callable=mock_open, read_data="key = 'value'")
    @patch("tomllib.load")
    def test_get_from_toml_unexpected_error(self, mock_load, mock_open):
        """
        Test case for handling an unexpected error while loading a TOML file.

        This test simulates a generic unexpected error during the TOML file parsing.
        It ensures that the function raises a `click.Abort` exception when any unexpected error occurs.

        Expected behavior:
        - A `click.Abort` exception is raised due to the unexpected error.
        - The file is opened with the correct path.
        - The `tomllib.load` function is called and raises the unexpected error.
        """
        # Simulate an unexpected error
        mock_load.side_effect = Exception("Unexpected error")

        # Test the function and assert that it raises click.Abort()
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.get_from_toml("/path/to/toml/file")

        # Check that the file was opened with the correct path
        mock_open.assert_called_once_with("/path/to/toml/file", "rb")

        # Ensure tomllib.load was called and it raised the unexpected error
        mock_load.assert_called_once_with(mock_open())

    def test_strip_version(self):
        """
        Test case for stripping version specifiers from package strings.

        This test checks that the function correctly strips version specifiers
        (e.g., `>=`, `<=`, `==`) from package names in version strings.

        Expected behavior:
        - The function returns the package name without the version specifier.
        """
        self.assertEqual(ckan_pilot.helpers.catalog.strip_version("package>=1.2.3"), "package")
        self.assertEqual(ckan_pilot.helpers.catalog.strip_version("package<=2.0.0"), "package")
        self.assertEqual(ckan_pilot.helpers.catalog.strip_version("package==1.0.0"), "package")
        self.assertEqual(ckan_pilot.helpers.catalog.strip_version("package"), "package")  # No version

    @patch("ckan_pilot.helpers.catalog.subproc.uv_remove_wrapper")
    @patch("ckan_pilot.helpers.catalog.strip_version")
    def test_remove_extension(self, mock_strip_version, mock_uv_remove):
        """
        Test that remove_extension properly removes an extension and its dependencies.
        """
        # Mock project object
        project = MagicMock()
        project.projectdir = "/fake/project/dir"

        extension_name = "ckanext-example"
        extension_requirements = ["dep1>=1.0", "dep2<=2.0", "ckanext-example==1.2"]
        group = "default"
        extra_flags = ["--force"]

        # Improved mock strip_version to handle all constraint symbols
        mock_strip_version.side_effect = (
            lambda x: x.split("<")[0].split(">")[0].split("=")[0].split("~")[0].split("!")[0]
        )

        # Call function
        ckan_pilot.helpers.catalog.remove_extension(
            project.projectdir, extension_name, extension_requirements, group, extra_flags
        )

        # Check extension removal
        mock_uv_remove.assert_any_call(
            "/fake/project/dir", group=extension_name, target=extension_name, extra_flags=extra_flags
        )

        # Check dependencies removal (excluding the extension itself)
        mock_uv_remove.assert_any_call("/fake/project/dir", group=group, target="dep1", extra_flags=extra_flags)
        mock_uv_remove.assert_any_call("/fake/project/dir", group=group, target="dep2", extra_flags=extra_flags)

        # Ensure extension itself isn't removed twice
        removed_targets = [call.kwargs["target"] for call in mock_uv_remove.call_args_list if "target" in call.kwargs]
        self.assertEqual(removed_targets.count(extension_name), 1)

    @patch("ckan_pilot.helpers.catalog.logger")
    @patch("ckan_pilot.helpers.catalog.os.system")
    def test_remove_editable_extension(self, mock_os_system, mock_logger):
        """
        Test that remove_editable_extension properly deletes the specified folder and logs the action.
        """
        extension_path = "/fake/path/to/extension"

        # Call the function
        ckan_pilot.helpers.catalog.remove_editable_extension(extension_path)

        # Check if os.system was called correctly
        mock_os_system.assert_called_once_with("rm -rf {0}".format(extension_path))

        # Ensure correct logging
        mock_logger.info.assert_called_once_with("Deleted editable extension: {0}".format(extension_path))

    def test_extract_url_success(self):
        """
        Test extracting URL components successfully.

        This test verifies that the `extract_url` method correctly parses an extension
        URL and version into its individual components.
        """
        # Test data
        extension_url = "https://github.com/ckan/ckanext-example"
        extension_version = "main"

        result = ckan_pilot.helpers.catalog.extract_url(extension_url, extension_version)
        expected = ("ckan", "ckanext-example", "main", "example")
        self.assertEqual(result, expected)

    def test_extract_url_failure(self):
        """
        Test the failure scenario when extracting URL components.

        This test verifies that the `extract_url` method raises an `IndexError`
        when given an invalid URL.
        """
        # Test data
        extension_url = "not-a-valid-url"
        extension_version = "not-a-valid-version"

        with self.assertRaises(IndexError):
            ckan_pilot.helpers.catalog.extract_url(extension_url, extension_version)

    def test_retrieve_requirements_url_success(self):
        """
        Test successful retrieval of a requirements URL for a given extension.

        This test ensures that the `retrieve_requirements_url` method correctly
        constructs the requirements URL based on the given extension URL and version.
        """
        # Test data
        extension_url = "https://github.com/ckan/ckanext-example"
        extension_version = "main"
        requirements_file = "requirements.txt"

        result = ckan_pilot.helpers.catalog.retrieve_requirements_url(
            extension_url, extension_version, requirements_file
        )
        expected = "https://raw.githubusercontent.com/ckan/ckanext-example/main/requirements.txt"
        self.assertEqual(result, expected)

    def test_retrieve_requirements_url_failure(self):
        """
        Test the failure scenario when retrieving the requirements URL.

        This test verifies that an invalid URL or version raises an `IndexError`.
        """
        # Test data
        extension_url = "not-a-valid-url"
        extension_version = "nota-valid-version"
        requirements_file = "not-a-valid-file"

        with self.assertRaises(IndexError):
            ckan_pilot.helpers.catalog.retrieve_requirements_url(extension_url, extension_version, requirements_file)

    def test_retrieve_declaration_url_success(self):
        """
        Test successful retrieval of a configuration declaration URL.

        This test verifies that the `retrieve_declaration_url` method correctly
        constructs the declaration URL based on the given extension URL and version.
        """
        extension_url = "https://github.com/ckan/ckanext-example"
        extension_version = "main"

        result = ckan_pilot.helpers.catalog.retrieve_declaration_url(extension_url, extension_version)
        expected = "https://raw.githubusercontent.com/ckan/ckanext-example/main/ckanext/example/config_declaration.yaml"
        self.assertEqual(result, expected)

    def test_retrieve_declaration_url_failure(self):
        """
        Test the failure scenario when retrieving the configuration declaration URL.

        This test verifies that an invalid URL or version raises an `IndexError`.
        """
        extension_url = "not-a-valid-url"
        extension_version = "not-a-valid-branch"

        with self.assertRaises(IndexError):
            ckan_pilot.helpers.catalog.retrieve_declaration_url(extension_url, extension_version)

    def test_parse_extension_options_success(self):
        """
        Test the successful parsing of extension options.

        This test ensures that the `parse_extension_options` method correctly formats
        and generates environment variable-like output for a list of extension options.
        """
        # Test data
        options = [
            {
                "description": "This is option 1",
                "example": "option_1_example",
                "default": "value1",
                "key": "OPTION_1",
                "required": True,
            },
            {
                "description": "This is option 2",
                "example": "option_2_example",
                "default": "value2",
                "key": "OPTION_2",
                "required": False,
            },
        ]

        expected_output = """
            # DESCRIPTION: This is option 1
            # EXAMPLE: option_1_example
            # DEFAULT: value1
            OPTION_1=value1

            # DESCRIPTION: This is option 2
            # EXAMPLE: option_2_example
            # DEFAULT: value2
            # OPTION_2=value2
        """

        result = ckan_pilot.helpers.catalog.parse_extension_options(options)

        # Strip leading/trailing whitespaces for a clean comparison
        self.assertEqual(normalize_indentation(result), normalize_indentation(expected_output))

    def test_parse_extension_options_failure(self):
        """
        Test the failure scenario when parsing extension options.

        This test simulates incorrect or missing fields in the extension options
        and ensures the method does not return the expected result.
        """
        # Test data with missing or incorrect fields
        options = [
            {
                "description": "This is option 1",
                "example": "option_1_example",
                "default": "value1",
                "key": "OPTION_1",
                # Missing 'required' field
            },
            {
                "description": "This is option 2",
                "example": "option_2_example",
                "default": "value2",
                "key": "OPTION_2",
                # Missing 'required' field
            },
        ]

        # Incorrect expected output (missing some fields or wrong format)
        expected_output = """
        # DESCRIPTION: This is option 1
        # EXAMPLE: option_1_example
        # DEFAULT: value1
        OPTION_1=value1

        # DESCRIPTION: This is option 2
        # EXAMPLE: option_2_example
        # DEFAULT: value2
        OPTION_2=value2  # Missing '#' comment in this test case for failure
        """

        result = ckan_pilot.helpers.catalog.parse_extension_options(options)

        # Assert that the result does not match the incorrect expected output
        self.assertNotEqual(normalize_indentation(result), normalize_indentation(expected_output))

    @patch("ckan_pilot.helpers.catalog.get")
    def test_fetch_extension_success(self, mock_get):
        """
        Test fetching metadata of an extension from the catalog.

        This test ensures that the `fetch_extension_metadata` method returns
        the correct metadata for a given extension.

        Args:
            mock_get (MagicMock): Mocked `ckan_pilot.helpers.catalog.get` function.
        """
        # Mock catalog data
        mock_get.return_value = mock_catalog_data

        catalog_source = "mock_catalog_source"
        extension_name = "ckanext-xloader"

        result = ckan_pilot.helpers.catalog.fetch_extension_metadata(catalog_source, extension_name)

        assert result["name"] == "ckanext-xloader"
        assert result["url"] == "https://github.com/ckan/ckanext-xloader"
        assert result["version"] == "2.1.0"

    @patch("ckan_pilot.helpers.catalog.get")
    def test_fetch_extension_not_found(self, mock_get):
        """
        Test fetching a non-existing extension.

        This test verifies that when trying to fetch metadata for an extension
        that does not exist in the catalog, the function returns None.
        """
        mock_get.return_value = mock_catalog_data

        catalog_source = "mock_catalog_source"
        extension_name = "nonexistent-extension"

        result = ckan_pilot.helpers.catalog.fetch_extension_metadata(catalog_source, extension_name)
        self.assertIsNone(result)

    def setUp(self):
        """
        Set up a mock project instance.

        This method sets up a mock project object with a predefined project directory
        to be used in various tests.
        """
        self.mock_project = MagicMock()
        self.mock_project.projectdir = "/mock/path/to/project"
        self.pyproject_path = "/mock/path/to/project/pyproject.toml"

    @patch("ckan_pilot.helpers.subproc.uv_add_wrapper")
    def test_install_extension_production_mode(self, mock_uv_add_wrapper):
        """
        Test installing an extension in production mode (non-dev).

        This test verifies that when installing an extension in production mode,
        the correct `uv_add_wrapper` function is called with the correct arguments.
        """
        ckan_pilot.helpers.catalog.install_extension(
            project_dir=self.mock_project.projectdir,
            extension_url="https://github.com/ckan/ckanext-xloader",
            extension_version="2.1.0",
            dev=False,
            extension_name="ckanext-xloader",
        )

        mock_uv_add_wrapper.assert_called_once_with(
            "/mock/path/to/project",
            group="ckanext-xloader",
            target="git+https://github.com/ckan/ckanext-xloader@2.1.0",
            requirements_file=None,
            editable=None,
        )

    @patch("ckan_pilot.helpers.subproc.uv_add_wrapper", side_effect=click.Abort)
    def test_install_extension_production_failure(self, mock_uv_add_wrapper):
        """
        Test failure when installing an extension in production mode.

        This test verifies that when an invalid extension URL is passed in production mode,
        the function raises a `click.Abort` exception.
        """
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.install_extension(
                project_dir=self.mock_project.projectdir,
                extension_url="bad_url",
                extension_version="no_version",
                dev=False,
                extension_name="ckanext-xloader",
            )

    @patch("ckan_pilot.helpers.subproc.git_clone_wrapper")
    @patch("ckan_pilot.helpers.subproc.uv_add_wrapper")
    def test_install_extension_dev_mode(self, mock_uv_add_wrapper, mock_git_clone_wrapper):
        """
        Test installing an extension in development mode.

        This test verifies that when installing an extension in development mode,
        both `git_clone_wrapper` and `uv_add_wrapper` are called with the appropriate arguments.
        """
        ckan_pilot.helpers.catalog.install_extension(
            project_dir=self.mock_project.projectdir,
            extension_url="https://github.com/ckan/ckanext-xloader",
            extension_version="2.1.0",
            dev=True,
            extension_name="ckanext-xloader",
        )

        mock_git_clone_wrapper.assert_called_once_with(
            version="2.1.0",
            git_url="https://github.com/ckan/ckanext-xloader",
            extra_flags=None,
            clone_target_dir="/mock/path/to/project/extensions/ckanext-xloader",
        )

        mock_uv_add_wrapper.assert_called_once_with(
            "/mock/path/to/project",
            group="ckanext-xloader",
            target="/mock/path/to/project/extensions/ckanext-xloader",
            editable=True,
            requirements_file=None,
        )

    @patch("ckan_pilot.helpers.subproc.git_clone_wrapper", side_effect=click.Abort)
    def test_install_extension_dev_clone_failure(self, mock_git_clone_wrapper):
        """
        Test failure when cloning an extension in development mode.

        This test verifies that when an invalid extension URL is passed in development mode,
        the function raises a `click.Abort` exception.
        """
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.install_extension(
                project_dir=self.mock_project.projectdir,
                extension_url="bad_url",
                extension_version="no_version",
                dev=True,
                extension_name="ckanext-xloader",
            )

    @patch(
        "ckan_pilot.helpers.catalog.retrieve_requirements_url",
        return_value="https://mocked-requirements-url/requirements.txt",
    )
    @patch("ckan_pilot.helpers.subproc.uv_add_wrapper")
    def test_install_requirements_success(self, mock_uv_add_wrapper, mock_retrieve_url):
        """
        Test successful installation of requirements.

        This test verifies that when a valid requirements file is provided,
        the function installs the required dependencies by calling `uv_add_wrapper` with
        the correct URL and file paths.
        """
        ckan_pilot.helpers.catalog.install_requirements_file(
            project_dir=self.mock_project.projectdir,
            extension_url="https://github.com/ckan/ckanext-xloader",
            extension_version="2.1.0",
            extension_name="ckanext-xloader",
            requirements_file="requirements.txt",
            editable=False,
        )

        mock_retrieve_url.assert_called_once_with(
            "https://github.com/ckan/ckanext-xloader", "2.1.0", "requirements.txt"
        )

        mock_uv_add_wrapper.assert_called_once_with(
            "/mock/path/to/project",
            group="ckanext-xloader",
            target=None,
            editable=False,
            requirements_file="https://mocked-requirements-url/requirements.txt",
        )

    @patch(
        "ckan_pilot.helpers.catalog.retrieve_requirements_url",
        return_value="https://mocked-requirements-url/requirements.txt",
    )
    @patch(
        "ckan_pilot.helpers.subproc.uv_add_wrapper",
        side_effect=sp.CalledProcessError(1, "mock_cmd", stderr=b"mocked error"),
    )
    def test_install_requirements_failure(self, mock_uv_add_wrapper, mock_retrieve_url):
        """
        Test failure when installing requirements with invalid data.

        This test verifies that when an invalid extension URL or requirements file is provided,
        the function raises a `click.Abort` exception.
        """
        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.install_requirements_file(
                project_dir=self.mock_project.projectdir,
                extension_url="https://invalid-url.com/invalid-repo",  # Invalid URL
                extension_version="999.999.999",  # Invalid version
                extension_name="bad_name",
                requirements_file="nonexistent_requirements.txt",  # Invalid requirements file
                editable=False,
            )

        mock_retrieve_url.assert_called_once()
        mock_uv_add_wrapper.assert_called_once()

    @patch(
        "ckan_pilot.helpers.catalog.retrieve_requirements_url",
        return_value="https://mocked-requirements-url/dev-requirements.txt",
    )
    @patch("ckan_pilot.helpers.subproc.uv_add_wrapper")
    def test_install_requirements_editable(self, mock_uv_add_wrapper, mock_retrieve_url):
        """
        Test installing requirements in editable mode (for dev).

        This test verifies that when installing a requirements file in editable mode,
        `uv_add_wrapper` is called with the correct editable flag and URL.
        """
        ckan_pilot.helpers.catalog.install_requirements_file(
            project_dir=self.mock_project.projectdir,
            extension_url="https://github.com/ckan/ckanext-xloader",
            extension_version="2.1.0",
            extension_name="ckanext-xloader",
            requirements_file="dev-requirements.txt",
            editable=True,
        )

        mock_retrieve_url.assert_called_once_with(
            "https://github.com/ckan/ckanext-xloader", "2.1.0", "dev-requirements.txt"
        )

        mock_uv_add_wrapper.assert_called_once_with(
            "/mock/path/to/project",
            group="ckanext-xloader",
            target=None,
            editable=True,
            requirements_file="https://mocked-requirements-url/dev-requirements.txt",
        )

    @patch("ckan_pilot.helpers.catalog.retrieve_declaration_url", return_value="https://mocked-declaration-url")
    @patch("ckan_pilot.helpers.catalog.get", return_value={"groups": [{"options": ["option1", "option2"]}]})
    @patch("ckan_pilot.helpers.catalog.parse_extension_options", return_value="mocked option output")
    @patch("builtins.open", new_callable=mock_open)
    def test_write_configuration_file_success(self, mock_open, mock_parse, mock_get, mock_retrieve):
        """
        Test writing a configuration file successfully.

        This test verifies that when a valid extension is passed, the function
        writes the correct configuration to the file.
        """
        mock_project = type("Project", (object,), {"projectdir": "/mock/projectdir"})
        mock_extension = {
            "name": "test_extension",
            "url": "https://mocked-url",
            "version": "1.0.0",
            "configuration": {"has_config_declaration": True, "options": []},
        }
        extension_name = "test_extension"

        ckan_pilot.helpers.catalog.write_configuration_file(mock_project.projectdir, mock_extension, extension_name)

        mock_open.assert_called_once_with("/mock/projectdir/config/test_extension.ini", "w")
        mock_open().write.assert_called_once_with(
            "### test_extension ###\n\n# === Generic Settings ===\nmocked option output"
        )

    @patch("ckan_pilot.helpers.catalog.retrieve_declaration_url", return_value="https://mocked-declaration-url")
    @patch("ckan_pilot.helpers.catalog.get", return_value={"groups": [{"options": ["option1", "option2"]}]})
    @patch("ckan_pilot.helpers.catalog.parse_extension_options", return_value="mocked option output")
    @patch("builtins.open", new_callable=mock_open)
    def test_write_configuration_file_no_declaration(self, mock_open, mock_parse, mock_get, mock_retrieve):
        """
        Test writing a configuration file when no declaration exists.

        This test verifies that when an extension has no configuration declaration,
        the function writes the correct configuration options to the file.
        """
        mock_project = type("Project", (object,), {"projectdir": "/mock/projectdir"})
        mock_extension = {
            "name": "test_extension",
            "url": "https://mocked-url",
            "version": "1.0.0",
            "configuration": {"has_config_declaration": False, "options": ["option1", "option2"]},
        }
        extension_name = "test_extension"

        ckan_pilot.helpers.catalog.write_configuration_file(mock_project.projectdir, mock_extension, extension_name)

        mock_open.assert_called_once_with("/mock/projectdir/config/test_extension.ini", "w")
        mock_open().write.assert_called_once_with("### test_extension ###\nmocked option output")

    @patch("ckan_pilot.helpers.catalog.retrieve_declaration_url", return_value="https://mocked-declaration-url")
    @patch("ckan_pilot.helpers.catalog.get", return_value={"groups": [{"options": ["option1", "option2"]}]})
    @patch("ckan_pilot.helpers.catalog.parse_extension_options", return_value="mocked option output")
    @patch("builtins.open", new_callable=mock_open)
    def test_write_configuration_file_error(self, mock_open, mock_parse, mock_get, mock_retrieve):
        """
        Test error handling when writing a configuration file.

        This test verifies that if an error occurs during the file write operation,
        a `click.Abort` exception is raised.
        """
        mock_project = type("Project", (object,), {"projectdir": "/mock/projectdir"})
        mock_extension = {
            "name": "test_extension",
            "url": "https://mocked-url",
            "version": "1.0.0",
            "configuration": {"has_config_declaration": True, "options": []},
        }
        extension_name = "test_extension"

        mock_open.side_effect = sp.CalledProcessError(1, "mock_cmd", stderr=b"mocked error")

        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.write_configuration_file(mock_project, mock_extension, extension_name)

    @patch("ckan_pilot.helpers.catalog.os.path.exists")
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_file_not_found(self, mock_get_from_toml, mock_exists):
        mock_exists.return_value = False
        with self.assertRaises(FileNotFoundError):
            ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path)
        mock_get_from_toml.assert_not_called()

    @patch("ckan_pilot.helpers.catalog.os.path.exists", return_value=True)
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_empty_dependency_groups(self, mock_get_from_toml, mock_exists):
        mock_get_from_toml.return_value = {}
        result = ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path)
        self.assertEqual(result, [])

    @patch("ckan_pilot.helpers.catalog.os.path.exists", return_value=True)
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_ignore_dev_groups_and_envvars(self, mock_get_from_toml, mock_exists):
        mock_get_from_toml.return_value = {
            "dependency-groups": {
                "ckanext-foo-dev": ["ckanext-bar-dev"],
                "ckanext-envvars": ["ckanext-envvars"],
                "ckanext-bar": ["ckanext-bar-dev", "ckanext-envvars"],
            }
        }
        result = ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path, "ckanext-envvars")
        self.assertEqual(result, ["ckanext-bar"])

    @patch("ckan_pilot.helpers.catalog.os.path.exists", return_value=True)
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_without_ignore_dev_groups_and_envvars(self, mock_get_from_toml, mock_exists):
        mock_get_from_toml.return_value = {
            "dependency-groups": {
                "ckanext-foo-dev": ["ckanext-bar-dev"],
                "ckanext-envvars": ["ckanext-envvars"],
                "ckanext-bar": ["ckanext-bar-dev", "ckanext-envvars"],
            }
        }
        result = ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path)
        self.assertEqual(result, ["ckanext-bar", "ckanext-envvars"])

    @patch("ckan_pilot.helpers.catalog.os.path.exists", return_value=True)
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_parse_multiple_extensions(self, mock_get_from_toml, mock_exists):
        mock_get_from_toml.return_value = {
            "dependency-groups": {
                "ckanext-foo": ["ckanext-bar", "ckanext-baz-dev"],
                "ckanext-qux": ["other-package", "ckanext-quux"],
                "random-group": ["ckanext-random"],
            }
        }
        result = ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path)
        expected = ["ckanext-bar", "ckanext-baz", "ckanext-foo", "ckanext-quux", "ckanext-qux", "ckanext-random"]
        self.assertEqual(result, expected)

    @patch("ckan_pilot.helpers.catalog.os.path.exists", return_value=True)
    @patch("ckan_pilot.helpers.catalog.get_from_toml")
    def test_duplicates_and_sorting(self, mock_get_from_toml, mock_exists):
        mock_get_from_toml.return_value = {
            "dependency-groups": {
                "ckanext-foo": ["ckanext-bar", "ckanext-bar-dev", "ckanext-foo"],
                "ckanext-bar": ["ckanext-foo-dev"],
            }
        }
        result = ckan_pilot.helpers.catalog.get_installed_extensions(self.pyproject_path)
        self.assertEqual(result, ["ckanext-bar", "ckanext-foo"])

    @patch("ckan_pilot.helpers.catalog.Project")
    @patch("ckan_pilot.helpers.catalog.get")
    @patch("ckan_pilot.helpers.catalog.parse_config_declaration")
    @patch("builtins.open", new_callable=mock_open)
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_write_core_ckan_configuration_success(
        self, mock_logger, mock_open_func, mock_parse, mock_get, mock_project
    ):
        # Setup mocks
        mock_project.return_value.ckan_version = "2.10"
        mock_project.return_value.projectdir = "/fake/project"

        mock_get.return_value = "yaml-content"
        mock_parse.return_value = "[ini]\nkey=value\n"

        # Call function
        ckan_pilot.helpers.catalog.write_core_ckan_configuration("/fake/project")

        # Assert correct URL requested
        mock_get.assert_called_once_with(
            "https://raw.githubusercontent.com/ckan/ckan/refs/tags/ckan-2.10/ckan/config/config_declaration.yaml"
        )

        # Assert parse called with the fetched yaml
        mock_parse.assert_called_once_with("yaml-content")

        # Assert file opened for writing core-ckan.ini
        mock_open_func.assert_called_once_with("/fake/project/config/core-ckan.ini", "w")

        # Assert write was called with correct content
        handle = mock_open_func()
        handle.write.assert_called_once_with("### CORE-CKAN ###\n[ini]\nkey=value\n")

        # Assert info log called
        mock_logger.info.assert_called_with("Core CKAN configuration written to /fake/project/config/core-ckan.ini")

    @patch("ckan_pilot.helpers.catalog.Project")
    @patch("ckan_pilot.helpers.catalog.get")
    @patch("ckan_pilot.helpers.catalog.parse_config_declaration")
    @patch("builtins.open", new_callable=mock_open)
    @patch("ckan_pilot.helpers.catalog.logger")
    def test_write_core_ckan_configuration_write_error(
        self, mock_logger, mock_open_func, mock_parse, mock_get, mock_project
    ):
        mock_project.return_value.ckan_version = "2.10"
        mock_project.return_value.projectdir = "/fake/project"
        mock_get.return_value = "yaml-content"
        mock_parse.return_value = "[ini]\nkey=value\n"

        err = sp.CalledProcessError(returncode=1, cmd="write")
        err.stderr = b"mocked error message"
        mock_open_func.side_effect = err

        with self.assertRaises(click.Abort):
            ckan_pilot.helpers.catalog.write_core_ckan_configuration("/fake/project")

        mock_logger.error.assert_called_with("Failed to write env file: /fake/project/config/core-ckan.ini")
        mock_logger.debug.assert_called_with("mocked error message")

    def write_ini(self, content):
        """Utility to create a temp INI file with content and return the file path."""
        tmp = NamedTemporaryFile("w+", delete=False)
        tmp.write(content)
        tmp.flush()
        tmp.close()
        self.addCleanup(lambda: os.remove(tmp.name))
        return tmp.name

    def read_ini(self, path):
        with open(path, "r", encoding="utf-8") as f:
            return f.read()

    def test_modify_existing_key(self):
        ini = self.write_ini("[app:main]\nckan.plugins = stats\n")
        ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments(ini, "app:main", "ckan.plugins", "envvars")
        content = self.read_ini(ini)
        self.assertIn("ckan.plugins = envvars", content)
        self.assertNotIn("ckan.plugins = stats", content)

    def test_insert_key_into_existing_section(self):
        ini = self.write_ini("[app:main]\n# some comment\n\n")
        ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments(ini, "app:main", "ckan.plugins", "envvars")
        content = self.read_ini(ini)
        self.assertIn("ckan.plugins = envvars", content)
        self.assertIn("[app:main]", content)
        self.assertIn("# some comment", content)

    def test_add_new_section_and_key(self):
        ini = self.write_ini("[other]\nkey = value\n")
        ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments(ini, "app:main", "ckan.plugins", "envvars")
        content = self.read_ini(ini)
        self.assertIn("[app:main]", content)
        self.assertIn("ckan.plugins = envvars", content)
        self.assertIn("[other]", content)

    def test_preserve_comments_and_formatting(self):
        ini = self.write_ini("[app:main]\n# a comment\nsome_key = some_value\n")
        ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments(ini, "app:main", "ckan.plugins", "envvars")
        content = self.read_ini(ini)
        self.assertIn("# a comment", content)
        self.assertIn("some_key = some_value", content)
        self.assertIn("ckan.plugins = envvars", content)

    def test_append_key_to_empty_section(self):
        ini = self.write_ini("[app:main]\n")
        ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments(ini, "app:main", "ckan.plugins", "envvars")
        content = self.read_ini(ini)
        self.assertIn("[app:main]", content)
        self.assertIn("ckan.plugins = envvars", content)

    def test_file_not_found_raises(self):
        with self.assertRaises(FileNotFoundError):
            ckan_pilot.helpers.catalog.set_ini_key_value_preserve_comments("/no/such/file.ini", "main", "x", "y")



import shutil
import subprocess as sp
import unittest
from unittest.mock import patch

from click.testing import CliRunner

from ckan_pilot.root import cli


class TestInitCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def test_init_success(self):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("initialized", result.output)
        self.assertEqual(result.exit_code, 0)

    @patch.dict("os.environ", {"PATH": "/some/path"}, clear=True)
    def test_init_tools_missing(self):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.1",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("not found", result.output)
        self.assertEqual(result.exit_code, 1)

    def test_init_permissions_error(self):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("check permissions", result.output)
        self.assertEqual(result.exit_code, 1)

    def test_init_value_error(self):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.1",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("Validation error", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.add_ckan_requirements",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_init_fail_add_requirements(self, mock_add_ckan_requirements):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("CKAN requirements to project failed", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch("click.confirm", return_value=True)
    @patch(
        "ckan_pilot.helpers.subproc.add_ckan_requirements",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_init_fail_add_requirements_cleanup_yes(self, mock_add_ckan_requirements, mock_confirm):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("CKAN requirements to project failed", result.output)
        self.assertIn("Project directory deleted.", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch("click.confirm", return_value=False)
    @patch(
        "ckan_pilot.helpers.subproc.setup_ckan_src",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_init_fail_setup_ckan_src_cleanup_no(self, mock_setup_ckan_src, mock_confirm):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("Add CKAN to project failed", result.output)
        self.assertIn("Project directory left intact.", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.add_ckan_dev_requirements",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_init_fail_add_dev_requirements(self, mock_add_ckan_dev_requirements):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("dev requirements to project failed", result.output)
        self.assertEqual(result.exit_code, 1)

    @patch(
        "ckan_pilot.helpers.subproc.setup_ckan_src",
        autospec=True,
        side_effect=sp.CalledProcessError(returncode=2, cmd=["bad"], stderr="Something bad happened".encode()),
    )
    def test_init_fail_setup_ckan_src(self, mock_add_ckan_src):
        test_init_options = [
            "init",
            "-n",
            "test",
            "--project-description",
            "Test description",
            "--ckan-version",
            "2.11.3",
            "-p",
            "3.10.16",
            "/tmp/ckan-pilot/testdir",
        ]
        result = self.runner.invoke(cli, test_init_options)
        self.assertIn("Add CKAN to project failed", result.output)
        self.assertEqual(result.exit_code, 1)

    def tearDown(self):
        shutil.rmtree("/tmp/ckan-pilot", ignore_errors=True)



import unittest
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from ckan_pilot.commands.stop import cli as stop_cli


class TestStopCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.mock_project = MagicMock()
        self.mock_project.projectdir = "/fake/path"
        self.mock_project.get_project_metadata.return_value = ("testproject", "1.0.0")

    @patch("ckan_pilot.commands.stop.subproc._compose_down")
    def test_stop_default(self, mock_compose_down):
        result = self.runner.invoke(stop_cli, obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_compose_down.assert_called_once_with("/fake/path", v=False, project_name="testproject")

    @patch("ckan_pilot.commands.stop.subproc._compose_down")
    def test_stop_with_v(self, mock_compose_down):
        result = self.runner.invoke(stop_cli, ["-v"], obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_compose_down.assert_called_once_with("/fake/path", v=True, project_name="testproject")



import unittest
from unittest.mock import MagicMock, mock_open, patch

from click.testing import CliRunner
from docker.errors import APIError, BuildError

from ckan_pilot.commands.dist import add_system_requirements, cli


class TestBuildCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.mock_project = MagicMock()
        self.mock_project.projectdir = "/fake/project"
        self.mock_project.get_project_metadata.return_value = ("my-ckan", "2.10")

    @patch("ckan_pilot.commands.dist.os.chmod")
    @patch("ckan_pilot.commands.dist.open", new_callable=mock_open)
    @patch("ckan_pilot.commands.dist.catalog.set_ini_key_value_preserve_comments")
    @patch("ckan_pilot.commands.dist.subproc.ckan_cli")
    @patch("ckan_pilot.commands.dist.docker.APIClient")
    @patch("ckan_pilot.commands.dist.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.commands.dist.catalog.get_installed_extensions")
    def test_build_dev_mode(  # noqa
        self,
        mock_get_extensions,
        mock_fetch_metadata,
        mock_docker_api,
        mock_ckan_cli,
        mock_set_ini,
        mock_open_file,
        mock_chmod,
    ):
        mock_set_ini.return_value = None
        mock_client = MagicMock()
        mock_docker_api.return_value = mock_client
        mock_client.build.return_value = iter([{"stream": "Successfully built\n"}])

        mock_get_extensions.return_value = ["ckanext-test"]
        mock_fetch_metadata.return_value = {
            "setup": {
                "distributions": {"alpine": {"required_system_packages": ["bash"]}},
                "init-config": ["pages initdb"],
            }
        }

        result = self.runner.invoke(
            cli,
            ["--catalog-source", "dummy-url", "build", "--dev"],
            obj=self.mock_project,
        )
        print(result)  # For debugging purposes
        assert result.exit_code == 0
        mock_ckan_cli.assert_called_once_with("/fake/project", generate_config=True, dev=True, prod=False)
        mock_open_file.assert_any_call("/fake/project/Dockerfile", "w")
        mock_open_file.assert_any_call("/fake/project/compose-dev/services/ckan/image/before-init.sh", "w")
        mock_client.build.assert_called_once_with(
            path="/fake/project", tag="my-ckan", rm=True, decode=True, nocache=True
        )  # noqa: E501

        # Ensure the INI modification was attempted
        mock_set_ini.assert_called_once_with("/fake/project/development.ini", "app:main", "ckan.plugins", "envvars")

    @patch("ckan_pilot.commands.dist.catalog.get_installed_extensions")
    @patch("ckan_pilot.commands.dist.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.commands.dist.docker.APIClient")
    @patch("ckan_pilot.commands.dist.open", new_callable=mock_open)
    @patch("ckan_pilot.commands.dist.os.chmod")
    @patch("ckan_pilot.commands.dist.subproc.ckan_cli")
    @patch("ckan_pilot.commands.dist.logger")
    @patch("ckan_pilot.commands.dist.catalog.set_ini_key_value_preserve_comments")
    def test_build_prod_mode(  # noqa
        self,
        mock_set_ini,
        mock_logger,
        mock_ckan_cli,
        mock_os_chmod,
        mock_open_file,
        mock_docker_api,
        mock_fetch_metadata,
        mock_get_extensions,
    ):
        mock_client = MagicMock()
        mock_docker_api.return_value = mock_client

        # Return stream-like logs from APIClient.build()
        mock_client.build.return_value = iter([{"stream": "Built prod image\n"}])

        mock_set_ini.return_value = None

        mock_get_extensions.return_value = ["ckanext-test"]
        mock_fetch_metadata.return_value = {
            "setup": {"distributions": {"alpine": {"required_system_packages": ["bash"]}}}
        }

        result = self.runner.invoke(cli, ["build"], obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_ckan_cli.assert_called_once_with("/fake/project", generate_config=True, dev=False, prod=True)
        mock_open_file.assert_any_call("/fake/project/Dockerfile", "w")
        mock_client.build.assert_called_once_with(
            path="/fake/project", tag="my-ckan", rm=True, decode=True, nocache=True
        )
        mock_logger.info.assert_any_call("Image 'my-ckan' built successfully.")

        # Ensure the INI modification was attempted
        mock_set_ini.assert_called_once_with("/fake/project/production.ini", "app:main", "ckan.plugins", "envvars")

    @patch("ckan_pilot.commands.dist.os.chmod")
    @patch("ckan_pilot.commands.dist.open", new_callable=mock_open)
    @patch("ckan_pilot.commands.dist.catalog.set_ini_key_value_preserve_comments")
    @patch("ckan_pilot.commands.dist.subproc.ckan_cli")
    @patch("ckan_pilot.commands.dist.docker.APIClient")
    @patch("ckan_pilot.commands.dist.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.commands.dist.catalog.get_installed_extensions")
    def test_build_error(  # noqa
        self,
        mock_get_extensions,
        mock_fetch_metadata,
        mock_docker_api,
        mock_ckan_cli,
        mock_set_ini,
        mock_open_file,
        mock_os_chmod,
    ):
        mock_client = MagicMock()
        mock_docker_api.return_value = mock_client
        mock_client.build.side_effect = APIError("Boom")

        mock_set_ini.return_value = None
        mock_get_extensions.return_value = ["ckanext-test"]
        mock_fetch_metadata.return_value = {
            "setup": {"distributions": {"alpine": {"required_system_packages": ["bash"]}}}
        }

        result = self.runner.invoke(cli, ["build"], obj=self.mock_project)

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("Docker build failed.", result.output)

    @patch("ckan_pilot.commands.dist.os.chmod")
    @patch("ckan_pilot.commands.dist.catalog.get_installed_extensions")
    @patch("ckan_pilot.commands.dist.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.commands.dist.docker.APIClient")
    @patch("ckan_pilot.commands.dist.open", new_callable=mock_open)
    @patch("ckan_pilot.commands.dist.subproc.ckan_cli")
    @patch("ckan_pilot.commands.dist.catalog.set_ini_key_value_preserve_comments")
    def test_build_error_builderror(  # noqa
        self,
        mock_set_ini,
        mock_ckan_cli,
        mock_open_file,
        mock_docker_api,
        mock_fetch_metadata,
        mock_get_extensions,
        mock_os_chmod,
    ):  # noqa
        mock_client = MagicMock()
        mock_docker_api.return_value = mock_client
        mock_set_ini.return_value = None

        mock_error = BuildError("failed", build_log=[{"stream": "step 1/5"}, {"stream": "step 2/5"}])
        mock_client.build.side_effect = mock_error

        mock_get_extensions.return_value = ["ckanext-test"]
        mock_fetch_metadata.return_value = {
            "setup": {"distributions": {"alpine": {"required_system_packages": ["bash"]}}}
        }

        result = self.runner.invoke(cli, ["build"], obj=self.mock_project)

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("Docker build failed.", result.output)

    @patch("ckan_pilot.commands.dist.os.chmod")
    @patch("ckan_pilot.commands.dist.catalog.get_installed_extensions")
    @patch("ckan_pilot.commands.dist.catalog.fetch_extension_metadata")
    @patch("ckan_pilot.commands.dist.docker.APIClient")
    @patch("ckan_pilot.commands.dist.open", new_callable=mock_open)
    @patch("ckan_pilot.commands.dist.subproc.ckan_cli")
    @patch("ckan_pilot.commands.dist.catalog.set_ini_key_value_preserve_comments")
    def test_build_with_custom_tag(  # noqa
        self,
        mock_set_ini,
        mock_ckan_cli,
        mock_open_file,
        mock_docker_api,
        mock_fetch_metadata,
        mock_get_extensions,
        mock_os_chmod,
    ):
        mock_client = MagicMock()
        mock_docker_api.return_value = mock_client

        # Simulate streaming build output
        mock_client.build.return_value = iter([{"stream": "OK\n"}])
        mock_set_ini.return_value = None

        # Fake catalog responses
        mock_get_extensions.return_value = ["ckanext-test"]
        mock_fetch_metadata.return_value = {"setup": {"distributions": {"alpine": {"required_system_packages": []}}}}

        result = self.runner.invoke(cli, ["build", "--tag", "custom-tag:latest"], obj=self.mock_project)
        print(result)

        self.assertEqual(result.exit_code, 0)
        mock_client.build.assert_called_once_with(
            path="/fake/project",
            tag="custom-tag:latest",
            rm=True,
            decode=True,
            nocache=True,
        )

    @patch("ckan_pilot.commands.dist.logger")
    def test_single_requirement(self, mock_logger):
        dockerfile_template = "FROM alpine\n{additional_system_packages}\nCMD echo Hello"
        expected = "FROM alpine\nRUN apk add --no-cache bash\nCMD echo Hello"
        result = add_system_requirements(["bash"], dockerfile_template)
        self.assertEqual(result, expected)
        mock_logger.info.assert_called_once_with("Writing system requirements in Dockerfile...")

    @patch("ckan_pilot.commands.dist.logger")
    def test_multiple_requirements(self, mock_logger):
        dockerfile_template = "FROM alpine\n{additional_system_packages}\nCMD echo Hello"
        expected = "FROM alpine\nRUN apk add --no-cache bash \\\n\t\tcurl \\\n\t\tgit\nCMD echo Hello"
        result = add_system_requirements(["bash", "curl", "git"], dockerfile_template)
        self.assertEqual(result, expected)
        mock_logger.info.assert_called_once_with("Writing system requirements in Dockerfile...")

    def test_placeholder_missing(self):
        # Should not fail even if the placeholder is not present
        dockerfile_template = "FROM alpine\nRUN something\nCMD echo Hello"
        result = add_system_requirements(["bash"], dockerfile_template)
        # Since placeholder is missing, nothing is changed
        self.assertEqual(result, dockerfile_template)



import unittest
from unittest.mock import patch

from click.testing import CliRunner

from ckan_pilot.commands.config import cli


class DummyProject:
    def __init__(self, projectdir):
        self.projectdir = projectdir


class TestConfigCLI(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.dummy_project = DummyProject("/fake/project/path")

    @patch("ckan_pilot.helpers.config.combine_ini_files_to_env")
    def test_generate_config_calls_combine(self, mock_combine):
        result = self.runner.invoke(cli, ["generate"], obj=self.dummy_project)

        self.assertEqual(result.exit_code, 0)
        mock_combine.assert_called_once_with(
            "/fake/project/path/config",
            "/fake/project/path/compose-dev/config/ckan/.env",
            [
                "CKAN__PLUGINS",
                "CKAN_MAX_UPLOAD_SIZE_MB",
                "CKAN_SQLALCHEMY_URL",
                "CKAN_DATASTORE_WRITE_URL",
                "CKAN_DATASTORE_READ_URL",
                "CKAN_SOLR_URL",
                "CKAN_REDIS_URL",
                "MAINTENANCE_MODE",
                "CKAN_SITE_URL",
                "CKAN_PORT",
                "CKAN__STORAGE_PATH",
                "CKAN__WEBASSETS__PATHCKAN_SYSADMIN_NAME",
                "CKAN_SYSADMIN_PASSWORD",
                "CKAN_SYSADMIN_EMAIL",
                "CKAN_SYSADMIN_NAME",
                "CKAN_SITE_ID",
                "CKAN__RESOURCE_FORMATS",
            ],
        )

    def test_cli_group_has_generate_command(self):
        self.assertIn("generate", cli.commands)



"""ckan-pilot"""

import logging
import os
import pathlib

import click
import click_log
from auto_click_auto import enable_click_shell_completion
from auto_click_auto.constants import ShellType

# Supported CKAN Versions
CKAN_VERSIONS = ["2.10.8", "2.11.3"]

# Supported tools
TOOLS_SUPPORTED = ["uv"]

# CKAN Github URLs
CKAN_GIT_URL = "https://github.com/ckan/ckan"
CKAN_GH_RAW_URL_PREFIX = "https://raw.githubusercontent.com/ckan/ckan/ckan-"

# Copier CKAN project template
CKAN_PROJECT_TEMPLATE = "https://github.com/keitaroinc/ckan-project-template"

# Add ckan-pilot prefix to all envvars
CONTEXT_SETTINGS = {"auto_envvar_prefix": "CKAN_PILOT"}

# Extension catalog url
CKAN_EXT_CATALOG_URL = "https://extensions.ckan.app/catalog.yaml"

# Directory to load subcommands from
cmd_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))

# Initialize and set click-log to use logger
logger = logging.getLogger(__name__)
click_log.basic_config(logger)


# Project class to hold all project options
class Project(object):
    def __init__(self, projectdir=None, ckan_version=None, python_version=None):
        self.projectdir = os.path.abspath(projectdir or ".")
        self.ckan_version = self.__get_ckan_version(projectdir)
        self.python_version = self.__get_python_version(projectdir)
        logger.debug("Project directory: {0}".format(self.projectdir))
        logger.debug("CKAN version: {0}".format(self.ckan_version))
        logger.debug("Python version: {0}".format(self.python_version))

    def __get_version_from_file(self, file):
        file_path = pathlib.Path(file)
        if file_path.exists():
            with open(file) as f:
                version = f.read().strip()
                f.close()
        else:
            version = None
        return version

    def __get_ckan_version(self, ckan_version):
        """Get the project CKAN version from projectdir/.ckan-version"""
        ckan_ver = self.__get_version_from_file(pathlib.Path(self.projectdir) / ".ckan-version")
        return ckan_ver

    def __get_python_version(self, python_version):
        """Get the project python version from projectdir/.python-version"""
        py_ver = self.__get_version_from_file(pathlib.Path(self.projectdir) / ".python-version")
        return py_ver

    def has_virtual_env(self):
        """Check if Project has a virtual env"""
        project_venv = pathlib.Path(self.projectdir) / ".venv"
        return project_venv.exists()

    def is_ckan_pilot_project(self):
        """Check if project_dir is a ckan-pilot project and if not abort"""
        ckan_pilot_file = pathlib.Path(self.projectdir) / ".ckan-pilot"
        if not ckan_pilot_file.exists():
            logger.error("{0} is not a valid ckan-pilot project, missing .ckan-pilot?".format(self.projectdir))
            raise click.Abort()

    def has_ckan_src(self):
        """Check if CKAN source code is present in projects virtual env"""
        ckan_src_dir = pathlib.Path(self.projectdir) / "src/ckan"
        return ckan_src_dir.exists()

    def get_project_metadata(self):
        from ckan_pilot.helpers.catalog import get_from_toml  # noqa

        name = get_from_toml(os.path.join(self.projectdir, "pyproject.toml"))["project"]["name"]
        version = get_from_toml(os.path.join(self.projectdir, "pyproject.toml"))["project"]["version"]
        return name, version


# ckan-pilot main CLI class and options
# Subcommands are loaded from the commands directory
class PckanCLI(click.MultiCommand):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(cmd_dir):
            if filename.endswith(".py") and filename != "__init__.py":
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(cmd_dir, name + ".py")
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)
        return ns["cli"]


# Define top level ckan-pilot CLI
@click.command(cls=PckanCLI, context_settings=CONTEXT_SETTINGS)
@click.option("-d", "--project-dir", help="CKAN project directory", required=False)
@click_log.simple_verbosity_option(logger, default="INFO", show_default=True)
@click.pass_context
def cli(ctx, project_dir):
    """ckan-pilot CLI for managing CKAN projects"""
    enable_click_shell_completion(program_name="ckan-pilot", shells={ShellType.BASH, ShellType.ZSH, ShellType.FISH})
    ctx.obj = Project(project_dir)



import configparser
import os

import click

from ckan_pilot.helpers import subproc
from ckan_pilot.root import logger


@click.group("git", short_help="Manages git repositories")
def cli():
    """Group of commands related to managing git."""


@cli.command("init")
@click.option(
    "--git-repo",
    help="Path to remote git repository where to init submodules",
)
@click.pass_obj
def init_command(project, git_repo):
    """
    Group of commands related to managing git.
    """

    project_dir_path = project.projectdir
    extensions_folder_path = project_dir_path + "/extensions"

    cmd = ["init", project_dir_path]

    subproc.git_wrapper(cmd, None)

    if git_repo:
        cmd = ["remote", "add", "origin", git_repo]
        try:
            subproc.git_wrapper(cmd, None)
        except Exception as e:
            logger.error(f"Failed to add remote repository: {e}")
            raise click.Abort() from e

    # List all items in the extensions directory
    all_items = os.listdir(extensions_folder_path)

    # Exclude specific folders
    excluded = {".git"}
    available_extensions = [
        item for item in all_items if item not in excluded and os.path.isdir(os.path.join(extensions_folder_path, item))
    ]

    # Parse all extensions and get remote URLs from .git config file
    for extension in available_extensions:
        extension_path = os.path.join(extensions_folder_path, extension)
        git_config_path = os.path.join(extension_path, ".git", "config")

        if not os.path.exists(git_config_path):
            logger.info(f"Skipping {extension}: no .git/config found")
            continue

        config = configparser.ConfigParser()
        config.read(git_config_path)

        try:
            remote_url = config['remote "origin"']["url"]
        except KeyError:
            logger.error(f"Skipping {extension}: no remote origin URL found")
            continue

        # Add all extensions as submodules
        try:
            subproc.git_wrapper(["submodule", "add", "--force", remote_url, extension], None)
            logger.info(f"Added submodule: {extension} -> {remote_url}")
        except Exception as e:
            logger.error(f"Failed to add submodule {extension}: {e}")
            raise click.Abort() from e



import click

from ckan_pilot.helpers import subproc
from ckan_pilot.root import logger


@click.command()
@click.option("--development", "--dev", is_flag=True, help="Run in development (foreground) mode.")
@click.option("-v", is_flag=True, help="Delete disks on cleanup")
@click.option("-b", is_flag=True, help="Start compose with --build")
@click.pass_obj
def cli(project, development, v, b):
    """Start the CKAN pilot project."""
    logger.info("Starting CKAN Pilot...")
    d = not development  # Detached unless --development is passed
    project_name = "{0}".format(project.get_project_metadata()[0])
    enable_bake = None

    if development:
        # If running in development mode, set the environment variable to enable bake
        enable_bake = {"COMPOSE_BAKE": "true"}

    logger.info("Starting ckan-pilot this may take a while...")

    subproc.start_compose(project.projectdir, d=d, v=v, b=b, project_name=project_name, bake=enable_bake)



from shutil import which

from ckan_pilot.root import TOOLS_SUPPORTED, logger


def check_tools(tools):
    """Check if tools are available"""
    for tool in tools:
        check_tool = which(tool)
        tool_found = True
        if check_tool:
            logger.debug("`{0}` found at: {1}".format(tool, check_tool))
        else:
            # Check if we support installation through ckan-pilot and if yes advise user to run tools install
            if tool in TOOLS_SUPPORTED:
                logger.error("`{0}` not found, please run `ckan-pilot tools install`".format(tool))
            else:
                logger.error(
                    "`{0}` not found, please install the necessary package providing it on your system".format(tool)
                )
            tool_found = False
    return tool_found



"""Manages CKAN extensions and catalogs.

Fetches, processes, and displays CKAN extension data from YAML files or URLs,
supporting multiple extension management commands.
"""

import sys

import click
from dotenv import dotenv_values
from rich import print
from rich.table import Table

from ckan_pilot.helpers import catalog
from ckan_pilot.root import CKAN_EXT_CATALOG_URL, logger


@click.group("catalog", short_help="Manages catalog")
@click.option(
    "--catalog-source",
    default=CKAN_EXT_CATALOG_URL,
    help="Path to CKAN extension catalog file or URL, default: https://extensions.ckan.app/catalog.yaml",
)
def cli(catalog_source):
    """
    Group of commands related to managing extensions.
    """
    # Store the global catalog in a context variable
    logger.debug("Using Catalog source: {0}".format(catalog))
    cli.catalog_source = catalog_source


@cli.command("list", short_help="List extensions from a YAML file")
@click.option("--installed", is_flag=True)
@click.option("--enabled", is_flag=True)
@click.pass_obj
def list_extensions(project, installed, enabled):
    """
    List available extensions in extension catalog.

    This command fetches the extensions data from a specified YAML source
    (either a file or a URL). If no source is provided, it uses the default URL
    defined in `CKAN_EXT_CATALOG_URL`. The extensions' name and description are
    displayed in a table.

    Returns:
        None
    """
    # Use the global catalog from the context
    catalog_source = cli.catalog_source

    # Determine source and load YAML
    data = catalog.get(catalog_source)

    # Set path of .env
    env_path = project.projectdir + "/compose-dev/config/ckan/.env"

    # Check if the data is empty or not loaded properly
    if not data or "extensions" not in data or not data["extensions"]:
        logger.error("No extensions found in catalog: {0}".format(catalog_source))
        return

    if enabled:
        # Read .env and store as object
        env_vars = dotenv_values(env_path)

        # Get all CKAN__PLUGINS
        current_plugins = env_vars.get("CKAN__PLUGINS", "")

        # Create a table to display extension information
        table = Table(show_header=True, show_lines=True)
        table.add_column("Name", header_style="bold", vertical="middle")
        table.add_column("Description", header_style="bold")

        # Build a lookup from catalog name → description
        extension_lookup = {ext.get("name"): ext.get("description", "").strip() for ext in data.get("extensions", [])}

        # Add each enabled plugin to the table
        for plugin in current_plugins.split():
            full_name = "ckanext-" + plugin
            description = extension_lookup.get(full_name, "") or "No description available"
            table.add_row(full_name, description)

        print(table)
        sys.exit()

    # Create a table to display extension information
    table = Table(show_header=True, show_lines=True)
    table.add_column("Name", header_style="bold", vertical="middle")
    table.add_column("Description", header_style="bold")

    if installed:
        installed_extensions_in_toml = catalog.get_installed_extensions(project.projectdir + "/pyproject.toml", None)
        # Loop through catalog and include only installed ones
        for ext in data.get("extensions", []):
            if ext.get("name") in installed_extensions_in_toml:
                name = "{}".format(ext.get("name"))
                description = "{}".format(ext.get("description", "").strip())
                table.add_row(name, description)
    else:
        # Add rows for each extension
        for ext in data.get("extensions", []):
            name = "{}".format(ext.get("name"))
            description = "{}".format(ext.get("description").strip())
            table.add_row(name, description)

    # Output table of extensions using print from rich
    print(table)


@cli.command("add", short_help="Add extensions to CKAN from catalog")
@click.option("--dev", is_flag=True)
@click.argument("extension_name", nargs=1)
@click.pass_obj
def extension_add(project, extension_name, dev):
    """
    Add a CKAN extension from the catalog.

    This command fetches the extension metadata from the catalog and installs it.
    If --dev is specified, the extension is cloned directly from its Git repository,
    instead of being installed as a packaged version.
    This allows developers to modify the source code locally while testing changes.
    The extension is installed in 'editable' mode, meaning changes take effect without needing reinstallation.

    Args:
        project (object): The CKAN project instance.
        extension_name (str): The name of the extension to add.
        dev (bool): Whether to install in development mode.
    """
    # Use the global yaml_source from the context
    catalog_source = cli.catalog_source
    extension = catalog.fetch_extension_metadata(catalog_source, extension_name)
    if not extension:
        logger.error(f"Extension '{extension_name}' not found in catalog {catalog_source}, skipping add.")
        return

    # Retrieve extension metadata
    extension_url = extension["url"]
    extension_version = extension["version"]
    # system_requirements = extension["setup"]["required_system_packages"]
    requirements = extension["setup"]["has_requirements"]
    requirements_dev = extension["setup"]["has_dev_requirements"]

    # Install the extension
    # If --dev is called it will install the extension as editable and clone it.
    catalog.install_extension(project.projectdir, extension_url, extension_version, dev, extension_name)

    # Install requirements if necessary
    if requirements:
        logger.info(f"Installing requirements for: {extension_name}")
        catalog.install_requirements_file(
            project.projectdir, extension_url, extension_version, extension_name, "requirements.txt", editable=False
        )

    if dev and requirements_dev:
        logger.info(f"Installing dev requirements for: {extension_name}")
        extension_name = extension_name + "-dev"
        catalog.install_requirements_file(
            project.projectdir, extension_url, extension_version, extension_name, "dev-requirements.txt", editable=False
        )

    # Install system requirements (Placeholder)
    # if system_requirements:
    #     logger.info("TODO: To be implemented")

    # Write configuration options to a .env file
    catalog.write_configuration_file(project.projectdir, extension, extension_name)



import subprocess as sp

import click

from ckan_pilot.helpers import checks, subproc
from ckan_pilot.root import logger


# Sync a CKAN project
@click.command("sync", short_help="Sync CKAN project")
@click.pass_obj
def cli(project):
    """Sync a CKAN project, its dependencies, extensions and config"""
    # Check if projectdir is a ckan-pilot project and if not Abort
    project.is_ckan_pilot_project()

    # Check that all required tools are in place otherwise Abort
    tools_required = ["uv", "git", "pg_config"]
    if not checks.check_tools(tools_required):
        raise click.Abort()

    # Check if we have a virtual environment
    if not project.has_virtual_env():
        try:
            subproc.create_virtual_env(project.projectdir)
        except sp.CalledProcessError as err:
            logger.error("Failed to create a virtual environment in {0}".format(project.projectdir))
            logger.debug(err.stderr.decode("utf-8"))
            raise click.Abort() from err

    # Check if we have CKAN source code in .venv and if not add it
    if not project.has_ckan_src():
        try:
            subproc.setup_ckan_src(project.projectdir, project.ckan_version)
        except sp.CalledProcessError as err:
            logger.error("Add CKAN to project failed")
            logger.debug(err.stderr.decode("utf-8"))
            raise click.Abort() from err

    # Sync ckan-pilot project
    try:
        subproc.sync_project(project.projectdir)
    except sp.CalledProcessError as err:
        logger.error("Failed to sync project")
        logger.debug(err.stderr.decode("utf-8"))
        raise click.Abort() from err

    logger.info("CKAN project synchronized at {0}".format(project.projectdir))



import os
import re

from ckan_pilot.root import logger


def transform_key(key: str) -> str:
    """
    Transform an INI-style config key into an environment variable style key.

    The transformation rules are:
    - Keys starting with 'ckanext.' become 'CKANEXT__' plus the uppercase remainder,
      with '.' replaced by '__'.
    - Keys starting with 'ckan.' become 'CKAN__' plus the uppercase remainder,
      with '.' replaced by '__'.
    - All other keys become 'CKAN___' plus the uppercase key,
      with '.' replaced by '__'.

    Args:
        key (str): The original configuration key.

    Returns:
        str: The transformed environment variable style key.
    """
    key_lower = key.lower()
    if key_lower.startswith("ckanext."):
        return "CKANEXT__" + key[8:].upper().replace(".", "__")
    elif key_lower.startswith("ckan."):
        return "CKAN__" + key[5:].upper().replace(".", "__")
    else:
        return "CKAN___" + key.upper().replace(".", "__")


def parse_ini_files(input_dir):  # noqa: PLR0912 PLR0915
    """
    Parse all `.ini` files in a directory, extracting settings grouped by sections.

    Sections are recognized by header lines matching `### SECTION ###`.
    Known sections (e.g. "### CORE-CKAN ###") are prioritized and preserved in order.
    Settings lines can be commented or uncommented. Comments like
    `# DESCRIPTION:`, `# EXAMPLE:`, and `# DEFAULT:` are ignored.
    Settings keys are transformed using `transform_key`.

    Args:
        input_dir (str): Path to the directory containing `.ini` files.

    Returns:
        tuple: (sections, section_order)
            - sections (dict): Mapping section header -> list of processed lines.
            - section_order (list): Ordered list of section headers as encountered.
    """
    setting_line_commented_regex = re.compile(r"^#\s*([A-Za-z0-9_.-]+)\s*=(.*)$")
    setting_line_uncommented_regex = re.compile(r"^([A-Za-z0-9_.-]+)\s*=(.*)$")
    section_header_regex = re.compile(r"^(###\s*[^#]+###)$")
    ignore_comment_prefixes = ("# DESCRIPTION:", "# EXAMPLE:", "# DEFAULT:")

    known_sections = {"### CORE-CKAN ###"}

    sections = {}
    section_order = []
    current_section = None

    for filename in sorted(os.listdir(input_dir)):
        if not filename.endswith(".ini"):
            continue
        filepath = os.path.join(input_dir, filename)
        with open(filepath, "r", encoding="utf-8") as infile:
            for line in infile:
                stripped_line = line.rstrip("\r\n")  # Remove newline chars only
                stripped = stripped_line.strip()

                section_match = section_header_regex.match(stripped)
                if section_match:
                    section_candidate = section_match.group(1)
                    if section_candidate in known_sections:
                        current_section = section_candidate
                        if current_section not in sections:
                            sections[current_section] = []
                            section_order.append(current_section)
                        sections[current_section].append(stripped_line + "\n")
                        continue
                    else:
                        # Treat unknown ### ... ### lines as normal lines inside current_section
                        pass

                if current_section is None:
                    if stripped == "" or any(stripped.startswith(prefix) for prefix in ignore_comment_prefixes):
                        continue
                    current_section = ""
                    if current_section not in sections:
                        sections[current_section] = []
                        section_order.append(current_section)

                if any(stripped.startswith(prefix) for prefix in ignore_comment_prefixes):
                    continue

                if stripped.startswith("#") and not setting_line_commented_regex.match(stripped):
                    sections[current_section].append(stripped_line + "\n")
                    continue

                match_commented = setting_line_commented_regex.match(stripped)
                if match_commented:
                    key = match_commented.group(1).strip()
                    val = match_commented.group(2).strip()
                    new_key = transform_key(key)
                    new_line = f"{new_key}={val}\n"
                    sections[current_section].append(new_line)
                    continue

                match_uncommented = setting_line_uncommented_regex.match(stripped)
                if match_uncommented:
                    key = match_uncommented.group(1).strip()
                    val = match_uncommented.group(2).strip()
                    new_key = transform_key(key)
                    new_line = f"{new_key}={val}\n"
                    sections[current_section].append(new_line)
                    continue

                sections[current_section].append(stripped_line + "\n")

    return sections, section_order


def write_env_file(sections, section_order, output_file):
    """
    Write combined environment variable settings to a `.env` file.

    The '### CORE-CKAN ###' section, if present, is always written first.
    Other sections are written preserving the original order.
    Blank lines are added before comment lines except at the file start.
    Unnamed (empty) sections are written last.

    Args:
        sections (dict): Mapping of section headers to lists of config lines.
        section_order (list): List defining the order of sections.
        output_file (str): Path to the output `.env` file to write.
    """
    core_section = None
    for sec in section_order:
        if sec.strip("# ").upper() == "CORE-CKAN":
            core_section = sec
            break

    try:
        with open(output_file, "w", encoding="utf-8") as outfile:
            first_line_written = False  # Track if anything has been written yet

            def write_section(lines):
                nonlocal first_line_written
                for line in lines:
                    stripped = line.strip()
                    if stripped == "":
                        continue

                    if stripped.startswith("#"):
                        # Add blank line before comment except if it's first line in the file
                        if first_line_written:
                            outfile.write("\n")
                        outfile.write(line.rstrip("\r\n") + "\n")
                        outfile.write("\n")
                    else:
                        outfile.write(line.rstrip("\r\n") + "\n")

                    first_line_written = True

            # Write CORE-CKAN first
            if core_section:
                logger.debug("Writing CORE-CKAN section first")
                write_section(sections[core_section])

            other_sections = [sec for sec in section_order if sec not in (core_section, "")]

            if core_section and other_sections:
                outfile.write("\n")

            for i, sec in enumerate(other_sections):
                logger.debug("Writing section: %s", sec)
                write_section(sections[sec])
                if i < len(other_sections) - 1:
                    outfile.write("\n")

            if "" in sections:
                if core_section or other_sections:
                    outfile.write("\n")
                logger.debug("Writing unnamed section")
                write_section(sections[""])

        logger.info("Output written to: %s", output_file)

    except Exception as e:
        logger.error("Error writing output file %s: %s", output_file, e)
        raise


def extract_env_variables(filepath, keys):
    """
    Extract specific environment variables from a file.

    Args:
        filepath (str): Path to the `.env` file.
        keys (list of str): Environment variable keys to extract.

    Returns:
        dict: A dictionary of {key: value} for found variables.
    """
    result = {}
    if not os.path.exists(filepath):
        return result

    keys_set = set(keys)

    with open(filepath, "r") as f:
        for line in f:
            stripped = line.strip()
            if "=" in stripped:
                k, v = stripped.split("=", 1)
                if k in keys_set:
                    result[k] = v
    return result


def combine_ini_files_to_env(input_dir, output_file, preserved_keys=None):
    """
    Combines .ini files into a single .env file, preserving specified keys.

    Args:
        input_dir (str): Directory containing .ini files.
        output_file (str): Output .env file path.
        preserved_keys (list of str): Environment variable keys to preserve.
    """
    logger.debug("Starting combination of ini files from %s into %s", input_dir, output_file)

    sections, section_order = parse_ini_files(input_dir)

    preserved_keys = preserved_keys or []
    preserved_values = extract_env_variables(output_file, preserved_keys)

    if preserved_values:
        logger.debug("Preserved variables: %s", preserved_values)
        if "### CORE-CKAN ###" not in section_order:
            section_order.insert(0, "### CORE-CKAN ###")
            sections.setdefault("### CORE-CKAN ###", [])

        core_section = sections.get("### CORE-CKAN ###", [])

        for key, value in preserved_values.items():
            # Remove existing key
            core_section = [line for line in core_section if not line.strip().startswith(f"{key}=")]
            # Append new preserved line
            core_section.append(f"{key}={value}\n")

        sections["### CORE-CKAN ###"] = core_section

    write_env_file(sections, section_order, output_file)



import unittest
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from ckan_pilot.commands.start import cli as start_cli


class TestStartCommand(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.mock_project = MagicMock()
        self.mock_project.projectdir = "/fake/path"
        self.mock_project.get_project_metadata.return_value = ("testproject",)

    @patch("ckan_pilot.commands.start.subproc.start_compose")
    def test_start_default(self, mock_start_compose):
        result = self.runner.invoke(start_cli, obj=self.mock_project)
        self.assertEqual(result.exit_code, 0)
        mock_start_compose.assert_called_once_with(
            "/fake/path", d=True, v=False, b=False, project_name="testproject", bake=None
        )

    @patch("ckan_pilot.commands.start.subproc.start_compose")
    def test_start_development_mode(self, mock_start_compose):
        result = self.runner.invoke(start_cli, ["--development"], obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_start_compose.assert_called_once_with(
            "/fake/path", d=False, v=False, b=False, project_name="testproject", bake={"COMPOSE_BAKE": "true"}
        )

    @patch("ckan_pilot.commands.start.subproc.start_compose")
    def test_start_with_all_flags(self, mock_start_compose):
        result = self.runner.invoke(start_cli, ["--development", "-v", "-b"], obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_start_compose.assert_called_once_with(
            "/fake/path", d=False, v=True, b=True, project_name="testproject", bake={"COMPOSE_BAKE": "true"}
        )

    @patch("ckan_pilot.commands.start.subproc.start_compose")
    def test_start_with_v_and_b_only(self, mock_start_compose):
        result = self.runner.invoke(start_cli, ["-v", "-b"], obj=self.mock_project)

        self.assertEqual(result.exit_code, 0)
        mock_start_compose.assert_called_once_with(
            "/fake/path", d=True, v=True, b=True, project_name="testproject", bake=None
        )
